/**
	$Id $

	Copyright (C) 2015 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.clients.web.plugins;

import net.sf.basedb.core.AnyToAny;
import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.Metadata;
import net.sf.basedb.core.Metadata.PropertyPath;

/**	
	Data loader implementation that load any-to-any links. This 
	implementation can also be used in standalone mode on list page.
	Use {@link #create(String, Item, String)} to create an instance
	and then {@link #getLinkedItem(DbControl, BasicItem)}.
	
	@author Nicklas
	@since 3.6
*/
public class AnyToAnyDataLoader
	implements DataLoader<BasicItem>
{

	/**
		Create a new data loader that load a value from an any-to-any link.
		
		@param linkName Name of the any-to-any link (required)
		@param linkType Target item type of the link (optional)
		@param linkProperty Property of the target item to load (optional)
	*/
	@SuppressWarnings("unchecked")
	public static AnyToAnyDataLoader create(String linkName, Item linkType, String linkProperty)
	{
		PropertyPath<BasicItem, ?> path = null;
		if (linkType != null || linkProperty != null)
		{
			path = (PropertyPath<BasicItem, ?>)Metadata.getInstance(linkType.getItemClass()).getPropertyPath(linkProperty, false);
		}
		return new AnyToAnyDataLoader(linkName, linkType, path);
	}
	
	
	private final String linkName;
	private final Item linkType;
	private final PropertyPath<BasicItem, ?> propertyPath;

	/**
		Load links with the specified name and target type. If not link
		type or property path is given, only the ID of the target item
		is returned.
		
		@param linkName Name of the any-to-any link (required)
		@param linkType Target item type of the link (optional)
		@param propertyPath Property of the target item to load (optional)
	*/
	public AnyToAnyDataLoader(String linkName, Item linkType, PropertyPath<BasicItem, ?> propertyPath)
	{
		this.linkName = linkName;
		this.linkType = linkType;
		this.propertyPath = propertyPath;
	}
	
	@Override
	public Object getData(DbControl dc, ExportedProperty exportedProperty, BasicItem item) 
	{
		Object value = null;
		BasicItem linked = getLinkedItem(dc, item);
		if (linked != null)
		{
			value = propertyPath == null ? linked.getId() : propertyPath.getValue(dc, linked);
		}
		
		return value;
	}

	/**
		Get the item that is linked to by this any-to-any link definition.
		@return The linked item or null if no item exists or if the item is not of
			the correct target type
	*/
	public BasicItem getLinkedItem(DbControl dc, BasicItem item)
	{
		BasicItem linked = null;
		try
		{
			AnyToAny any = AnyToAny.getByName(dc, item, linkName);
			if (linkType == null || linkType == any.getToType())
			{
				linked = any.getTo();
			}
		}
		catch (ItemNotFoundException ex)
		{}

		return linked;
	}
}
