/*
	$Id: Client.java 7403 2017-10-04 07:38:59Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import java.util.Date;

import net.sf.basedb.core.data.ClientData;
import net.sf.basedb.core.data.HelpData;
import net.sf.basedb.core.hibernate.TypeWrapper;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;

/**
	This class represents a registered client application. Registering a
	client application has some benefits:
	<ul>
	<li>It is possible to create and store settings
	<li>It is possible to set permissions for the usage of the application
	</ul>

	@author Nicklas
	@version 2.0
	@see ClientDefaultSetting
	@see UserClientSetting
*/
public class Client
	extends CommonItem
	implements Registered
{
	/**
		The type of item represented by this class.
		@see Item#CLIENT
		@see #getType()
	*/
	public static final Item TYPE = Item.CLIENT;
	
	/**
		The maximum length of the external id variable that can be stored in the
		database. Check the length against this value before calling the
		{@link #setExternalId(String)} method to avoid
		exceptions.
	*/
	public static final int MAX_EXTERNAL_ID_LENGTH = ClientData.MAX_EXTERNAL_ID_LENGTH;

	/**
		Create a new <code>Client</code> item.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access
		@param externalId The external id of the new item
		@return The new <code>Role</code> item
		@throws BaseException If there is an error
	*/
	public static Client getNew(DbControl dc, String externalId)
		throws BaseException
	{
		Client c = dc.newItem(Client.class);
		c.setName("New client");
		c.setExternalId(externalId);
		c.getData().setEntryDate(new Date());
		return c;
	}

	/**
		Get a <code>Client</code> item when you know the ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param id The ID of the item to load
		@return The <code>Client</code> item
		@throws ItemNotFoundException If an item with the specified 
			ID is not found
		@throws PermissionDeniedException If the logged in user doesn't 
			have {@link Permission#READ} permission to the item
		@throws BaseException If there is another error
	*/
	public static Client getById(DbControl dc, int id)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		Client c = dc.loadItem(Client.class, id);
		if (c == null) throw new ItemNotFoundException("Client[id="+id+"]");
		return c;
	}

	/**
		Get a <code>Client</code> item when you know the external ID.

		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param externalId The external id of the item to load
		@return The <code>Client</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ READ} permission for the item
		@throws BaseException If there is another error
	*/
	public static Client getByExternalId(DbControl dc, String externalId)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (externalId != null) externalId = externalId.trim();
		org.hibernate.query.Query<ClientData> query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
			"GET_CLIENT_FOR_EXTERNAL_ID", ClientData.class);
		/*
			SELECT cli
			FROM ClientData cli
			WHERE cli.externalId = :externalId
		*/
		query.setParameter("externalId", externalId, TypeWrapper.H_STRING);
		Client c = dc.getItem(Client.class, HibernateUtil.loadData(query));
		if (c == null) throw new ItemNotFoundException("Client[externalId="+externalId+"]");
		return c;
	}

	/**
		Get a new query object for this class.
		@return An {@link ItemQuery} object
	*/
	public static ItemQuery<Client> getQuery()
	{
		return new ItemQuery<Client>(Client.class);
	}

	Client(ClientData clientData)
	{
		super(clientData);
	}
	
	@Override
	ClientData getData()
	{
		return (ClientData)getBasicData();
	}

	/*
		From the Identifiable interface
		-------------------------------------------
	*/
	@Override
	public Item getType()
	{
		return TYPE;
	}
	// -------------------------------------------
	/*
		From the Registered interface
		-------------------------------------------
	*/
	@Override
	public Date getEntryDate()
	{
		return DateUtil.copy(getData().getEntryDate());
	}
	@Override
	public void setEntryDate(Date entryDate)
	{
		if (isInDatabase()) 
		{
			throw new PermissionDeniedException("Not allowed to change entryDate of " + this);
		}
		getData().setEntryDate(DateUtil.setDateOrToday(entryDate));
	}
	// -------------------------------------------

	/**
		Get the external id for this <code>Client</code> item.
		@return A string with the external id of this item
	*/
	public String getExternalId()
	{
		return getData().getExternalId();
	}
	/**
		Set the external id for this <code>Client</code> item. The value
		cannot be null and must not be longer than the value specified by
		the {@link #MAX_EXTERNAL_ID_LENGTH} constant.
		@param externalId The new external id for this item
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the new value is null or longer than
			{@link #MAX_EXTERNAL_ID_LENGTH}
	*/
	public void setExternalId(String externalId)
		throws PermissionDeniedException, InvalidDataException
	{
		checkPermission(Permission.WRITE);
		getData().setExternalId(StringUtil.setNotNullString(externalId, "externalId", MAX_EXTERNAL_ID_LENGTH));
	}

	/**
		This flag indicates if the registered client application supports 
		device verification or not. 
		@return TRUE if it is supported, FALSE if not
		@since 3.12
	*/
	public boolean getSupportsDeviceVerification()
	{
		return getData().getSupportsDeviceVerification();
	}
	public void setSupportsDeviceVerification(boolean supportsDeviceVerification)
	{
		checkPermission(Permission.WRITE);
		getData().setSupportsDeviceVerification(supportsDeviceVerification);
	}

	
	/**
		Create a new default setting for this client.
		@param name The name of the setting
		@param value The value of the setting
		@return The new <code>ClientDefaultSetting</code> object
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the name or value is null or 
			longer than what is allowed
		@throws BaseException If there is another error
	*/
	public ClientDefaultSetting newDefaultSetting(String name, String value)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		return ClientDefaultSetting.getNew(getDbControl(), this, name, value);
	}
	
	/**
		Get the default setting with the specified name.
		@param name The name of the setting
		@throws ItemNotFoundException If a setting with the specified
			name isn't found
		@throws InvalidDataException If name is null
		@throws BaseException If there is another error
	*/
	public ClientDefaultSetting getDefaultSetting(String name)
		throws ItemNotFoundException, InvalidDataException, BaseException
	{
		if (name == null) throw new InvalidUseOfNullException("name");
		name = name.trim();
		ItemQuery<ClientDefaultSetting> query = getDefaultSettings();
		query.restrict(
			Restrictions.eq(
				Hql.property("name"), 
				Expressions.parameter("name")
			)
		);
		query.setParameter("name", name, Type.STRING);
		ItemResultList<ClientDefaultSetting> l = query.list(getDbControl());
		if (l.size() == 0)
		{
			throw new ItemNotFoundException("ClientDefaultSetting[name="+name+"]");
		}
		return l.get(0);
	}
	
	/**
		Get a query that will return all default settings for this client.
		@return An {@link ItemQuery} object
	 	@throws BaseException If getting the query fails. 
	*/
	public ItemQuery<ClientDefaultSetting> getDefaultSettings()
		throws BaseException
	{
		return ClientDefaultSetting.getQuery(this);
	}

	/**
		Create a new help text for this client.
		@param externalId The external ID of the help text
		@return The new <code>Help</code> object
		@throws PermissionDeniedException If the logged in user doesn't have
			write permission
		@throws InvalidDataException If the exernal ID is null or 
			longer than what is allowed
		@throws BaseException If there is another error
	*/
	public Help newHelp(String externalId)
		throws PermissionDeniedException, InvalidDataException, BaseException
	{
		checkPermission(Permission.WRITE);
		return Help.getNew(getDbControl(), this, externalId);
	}
	
	/**
		Get a query that will return all help texts for this client.
		@return An {@link ItemQuery} object
	*/
	public ItemQuery<Help> getHelp()
	{
		return Help.getQuery(this);
	}
	
	/**
		Get a <code>Help</code> item when you know the external ID.
	
		@param dc The <code>DbControl</code> which will be used for
			permission checking and database access.
		@param externalId The external id of the item to load
		@param create TRUE if a new Help should be created if one doesn't
			exists, FALSE to throw an exception.
		@return The <code>Help</code> item
		@throws ItemNotFoundException If an item with the specified ID is not found.
		@throws PermissionDeniedException If the logged in user doesn't have
			{@link Permission#READ READ} permission for the item.
		@throws BaseException If there is another error.
	*/
	public Help getHelpByExternalId(DbControl dc, String externalId, boolean create)
		throws ItemNotFoundException, PermissionDeniedException, BaseException
	{
		if (externalId != null) externalId = externalId.trim();
		org.hibernate.query.Query<HelpData> query = HibernateUtil.getPredefinedQuery(dc.getHibernateSession(), 
			"GET_HELP_FOR_EXTERNAL_ID", HelpData.class);
		/*
			SELECT hlp
			FROM HelpData hlp
			WHERE hlp.externalId = :externalId AND hlp.client = :client
		*/
		query.setParameter("externalId", externalId, TypeWrapper.H_STRING);
		query.setParameter("client", getId(), TypeWrapper.H_INTEGER);
		Help h = dc.getItem(Help.class, HibernateUtil.loadData(query));
		if (h == null)
		{
			if(!create)
			{
				throw new ItemNotFoundException("Help[externalId="+externalId+"]");
			}
			h = newHelp(externalId);			
		}			
		return h;
	}
}



