/*
	$Id: DynamicQuery.java 7433 2017-12-08 13:18:12Z nicklas $
	
	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Dynamic;

/**
	Represents a query against the dynamic database. A single
	query object may query against data in one data cube at a time. Most of
	the times however, this is limited to data in a single bioassayset.
	
	A query usually has the {@link VirtualTable#SPOT} or {@link VirtualTable#POSITION}
	as the root table for the query. Use factory methods in the {@link Dynamic} 
	class to join other tables and create restrictions based on the data in those tables.
	<p>
	This type of query is executed by the {@link #iterate(DbControl)} method
	and returns the result as a {@link DynamicResultIterator}.
	
	@author Nicklas
	@version 2.0
	@see BioAssaySet#getSpotData()
	@see BioAssaySet#getPositionData()
	@see ExtraValue#getValues()
	@base.modified $Date: 2017-12-08 14:18:12 +0100 (fr, 08 dec 2017) $
*/
public abstract class DynamicQuery
	extends AbstractSqlQuery
	implements ClonableReporterQuery
{

	private final DataCube cube;
	private final VirtualDb virtualDb;
	private final VirtualTable rootTable;
	private final IntensityTransform transform;
	private final boolean hasClonedReporters;
	private boolean useClonedReporters;

	/**
		Create a new dynamic query.
		@param cube The cube to query against
		@param rootTable The root table of the query
	*/
	DynamicQuery(DataCube cube, VirtualTable rootTable, IntensityTransform transform)
	{
		super(rootTable.getQualifiedTableName(cube.getVirtualDb()));
		this.cube = cube;
		this.virtualDb = cube.getVirtualDb();
		this.rootTable = rootTable;
		this.transform = transform;
		this.hasClonedReporters = virtualDb.hasClonedReporters();
		this.useClonedReporters = hasClonedReporters;
		restrictPermanent(
			new DynamicRestriction(VirtualColumn.CUBE, cube.getCubeNo())
		);

	}
	
	/**
		Create a new non-restricted dynamic query.
		@param vdb The virtual database to use
		@param rootTable The root table of the query
		@since 3.1
	 */
	DynamicQuery(VirtualDb vdb, VirtualTable rootTable)
	{
		super(rootTable.getQualifiedTableName(vdb));
		this.cube = null;
		this.virtualDb = vdb;
		this.hasClonedReporters = vdb.hasClonedReporters();
		this.rootTable = rootTable;
		this.transform = null;
	}
	
	/*
		From the Query interface
		-------------------------------------------
	*/
	/**
		The alias of the item that is returned by this query.
		@see Item#getAlias()
	*/
	@Override
	public String getRootAlias()
	{
		return HibernateUtil.quote(rootTable.getAlias());
	}
	// -------------------------------------------
	
	/*
		From the ClonableReporterQuery interface
		----------------------------------------
	*/
	/**
		Check if the virtual database that this query is using has 
		cloned reporter annotations or not.
		@since 3.1
	*/
	@Override
	public boolean hasClonedReporters()
	{
		return hasClonedReporters;
	}
	
	/**
		Checks if this query is using cloned reporter or not by default.
		The default setting is to use cloned reporters if they are available
		in the experiment.
		
		@return TRUE if this query is using cloned reporters, FALSE if not
		@since 3.1
	*/
	@Override
	public boolean useClonedReporters()
	{
		return useClonedReporters;
	}
	
	/**
		Set if cloned reporters should be used by default or not.
		@param useClonedReporters TRUE to use cloned reporters by default, FALSE
			to use the master reporters
		@since 3.1
	*/
	@Override
	public void setUseClonedReporters(boolean useClonedReporters)
	{
		this.useClonedReporters = hasClonedReporters && useClonedReporters;
	}
	
	/**
		Get the virtual database that this query is using.
		@return A virtual database item
		@since 3.1
	*/
	@Override
	public VirtualDb getVirtualDb()
	{
		return virtualDb;
	}
	// -------------------------------------------
	
	DataCube getDataCube()
	{
		return cube;
	}

	VirtualTable getRootTable()
	{
		return rootTable;
	}
	
	/**
		Get the intensity transform that was used to store spot intensities
		for the root bioassay set in this query.
		@since 2.12
	*/
	public IntensityTransform getIntensityTransform()
	{
		return transform;
	}
	
}
