/*
	$Id: ItemContext.java 7433 2017-12-08 13:18:12Z nicklas $

	Copyright (C) 2005, 2006 Jari Häkkinen, Nicklas Nordborg
	Copyright (C) 2007 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core;

import net.sf.basedb.core.query.Annotations;
import net.sf.basedb.core.query.EntityQuery;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Orders;
import net.sf.basedb.core.query.Order;
import net.sf.basedb.core.query.Query;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.core.query.Restriction;
import net.sf.basedb.core.query.Expression;
import net.sf.basedb.core.query.Dynamic;
import net.sf.basedb.core.query.JoinType;
import net.sf.basedb.core.query.Select;
import net.sf.basedb.core.query.Selects;
import net.sf.basedb.core.query.SqlQuery;
import net.sf.basedb.core.data.ContextData;
import net.sf.basedb.core.data.ItemListSyncFilterData;
import net.sf.basedb.core.data.PropertyFilterData;
import net.sf.basedb.core.data.UserData;
import net.sf.basedb.core.data.ClientData;
import net.sf.basedb.util.StringLengthComparator;
import net.sf.basedb.util.encode.EncodeUtil;
import net.sf.basedb.util.encode.TabCrLfEncoderDecoder;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.jep.Jep;
import net.sf.basedb.util.jep.JepFunction;

import java.util.ArrayList;
import java.util.Arrays;
import java.util.Collection;
import java.util.Iterator;
import java.util.Map;
import java.util.HashMap;
import java.util.Set;
import java.util.HashSet;
import java.util.EnumSet;
import java.util.List;
import java.util.TreeSet;

/**
	This item is a helper class for client applications when they need to store
	information about the current context for plugins that needs to find out the
	current context of the running client application. It is optional to use this 
	class and it is recommended that plugins and client applications that expects 
	to find context information also can handle situations without any context 
	information. Some of the settings in this objects are saved in the database,
	some not.
	<p>
	Typical example for using this class are:
	<p>
	
	<b>A client application storing list/table settings</b><br>
	<pre class="code">
SessionControl sc = ...
ItemContext current = sc.getCurrentContext(Item.SAMPLE);
current.setRowsPerPage(50);
current.setPage(0);
current.setSortProperty("name");
current.setSortDirection(SortDirection.DESC);
current.setPropertyFilter(
      new PropertyFilter("name", Operator.LIKE, "Nicklas%", Type.STRING)
   );
ItemQuery query = Sample.getQuery();
current.configureQuery(query, true);
// The query will now return the 50 first samples with a
// a name starting with 'Nicklas' sorted in descending order
</pre>
	
	<p>
	
	<b>A client application using filter information when creating new items</b><br>
	<pre class="code">
SessionControl sc = ...
DbControl dc = sc.newDbControl();
ItemContext current = sc.getCurrentContext(Item.SAMPLE);
Sample sample = Sample.getNew(dc);
if (current.getPropertyValue("name") != null)
{
   sample.setName(current.getPropertyValue("name"));
}
// The name is now 'Nicklas%'
</pre>

	<p>
	<b>Plugin finds out which items in a list has been selected</b><br>
	<pre class="code">
SessionControl sc = ...
ItemContext current = sc.getCurrentContext(Item.SAMPLE);
ItemQuery query = Sample.getQuery();
Set selected = current.getSelected();
if (selected != null &amp;&amp; selected.size() &gt; 0)
{
   query.restrict(
         Restrictions.in(Hql.property("id"), 
         Expressions.parameter("selected", selected, Type.INT))
      );
)
// The query will now return the selected items only
</pre>
	
	<p>
	<b>An import plugin asks client application for a file with a specific file type</b><br>
	<pre class="code">
public RequestInformation getRequestInformation(GuiContext context, String command)
   throws BaseException
{
   RequestInformation requestInformation = null;
   if (command.equals(Request.COMMAND_CONFIGURE_PLUGIN))
   {
      requestInformation = getConfigurePluginParameters(context);
   }
   else if (command.equals(Request.COMMAND_CONFIGURE_JOB))
   {
      requestInformation = getConfigureJobParameters(context);
      ItemContext fileContext = sc.getCurrentContext(Item.FILE);
      fileContext.setPropertyFilter(FileType.getPropertyFilter(FileType.REPORTER));
   }
   return requestInformation;
}
</pre>

	@author Nicklas
	@version 2.0
	@base.modified $Date: 2017-12-08 14:18:12 +0100 (fr, 08 dec 2017) $
	@see SessionControl#getCurrentContext(Item)
	@see SessionControl#saveCurrentContextAs(Item, String, boolean, boolean)
*/
public class ItemContext
{

	/**
		This is the name of the default context which is automatically saved
		when a user is logging out. A client application shouldn't use this name
		for other named contexts that it wants to reload in the future.
	*/
	public static final String DEFAULT_NAME = "- default -";
	
	private final String name;
	private final Item itemType;
	private final String subContext;
	private int contextId;
	private int itemId;
	private Set<Integer> selected;
	private boolean noPaging = false;
	private int rowsPerPage;
	private int page;
	private int numFilterRows = 1;
	private String sortProperty = null;
	private String message;
	private SortDirection sortDirection = SortDirection.ASC;
	private Set<Include> include;
	private Permission permission;
	private Map<String, PropertyFilterPair> propertyFilters;
	private Map<String, String> settings;
	private Map<String, Object> objects;
	private Query query;
	private Filter<String> propertyInspector;
	
	/**
		Create a new context.
		@param sortProperty The property used to sort a list
		@param sortDirection If the items should be sorted in ascending or descending order
		@param rowsPerPage The number of rows to display on a single page
		@param page The page number of the currently displayed page, starting at 0
	*/
	public ItemContext(String sortProperty, SortDirection sortDirection, int rowsPerPage, int page)
	{
		this.name = null;
		this.itemType = null;
		this.subContext = null;
		this.sortProperty = sortProperty;
		this.sortDirection = sortDirection;
		this.rowsPerPage = rowsPerPage;
		this.page = page;
	}
	
	/**
		Create a new empty context.
	*/
	ItemContext(Item itemType, String subContext, String name)
	{
		this.itemType = itemType;
		this.subContext = subContext;
		this.name = name;
		this.rowsPerPage = 30;
		this.page = 0;
		this.include = EnumSet.of(Include.MINE);
	}
	
	/**
		Create a new context from a loaded database object.
	*/
	ItemContext(ContextData context)
	{
		this.itemType = Item.fromValue(context.getItemType());
		this.subContext = context.getSubContext();
		this.name = context.getName();
		this.contextId = context.getId();
		this.itemId = context.getItemId();
		this.rowsPerPage = context.getRowsPerPage();
		this.page = context.getPage();
		this.sortProperty = context.getSortProperty();
		this.sortDirection = SortDirection.fromValue(context.getSortDirection());
		this.include = Include.fromInt(context.getInclude());
		initPropertyFilters(context.getPropertyFilters());
		if (context.getSettings().size() != 0)
		{
			settings = new HashMap<String, String>(context.getSettings());
		}
	}
	
	/**
		Initialize the context from an item list synchronization filter.
		Note that this creates a special context used for implementing
		the query functionality of a {@link SyncFilter}. It should
		not be used for representing context in a GUI. To do this start
		with a regular context and the use {@link #copyFromSyncFilter(ItemListSyncFilterData)}
		instead.
		@since 3.5
	*/
	ItemContext(Item itemType, ItemListSyncFilterData syncFilter)
	{
		this.itemType = itemType;
		this.subContext = null;
		this.name = null;
		this.noPaging = true;
		copyFromSyncFilter(syncFilter);
	}
	
	/**
		Create a new database context object from this context. 
		@param name The name to use, or null if the current name should be used
		@param user The user that should own the context object
		@param client The client application currently in use
		@param isPublic If the context should be public or not
	*/
	ContextData getData(String name, UserData user, ClientData client, boolean isPublic)
	{
		ContextData context = new ContextData();
		context.setUser(user);
		context.setClient(client);
		context.setPublic(isPublic);
		context.setName(name == null ? this.name : name);
		context.setSubContext(subContext == null ? "" : subContext);
		context.setItemType(itemType.getValue());
		context.setItemId(itemId);
		context.setRowsPerPage(rowsPerPage);
		context.setPage(page);
		context.setSortProperty(sortProperty);
		if (sortDirection != null) context.setSortDirection(sortDirection.getValue());
		if (include != null) context.setInclude(Include.toInt(include));
		if (propertyFilters != null)
		{
			storePropertyFilters(context.getPropertyFilters());
		}
		if (settings != null)
		{
			context.getSettings().putAll(settings);
		}
		return context;
	}

	/**
		Create a new item list synchronization filter from this context. Basically the
		same as {@link #getData(String, UserData, ClientData, boolean)} but only
		keep filter-related settings.
		@since 3.5
	*/
	void copyToSyncFilter(ItemListSyncFilterData syncFilter)
	{
		syncFilter.setSourceItemType(getItemType().getValue());
		if (include != null) syncFilter.setInclude(Include.toInt(include));
		syncFilter.getPropertyFilters().clear();
		if (propertyFilters != null)
		{
			storePropertyFilters(syncFilter.getPropertyFilters());
		}
	}
	
	void copyFromSyncFilter(ItemListSyncFilterData syncFilter)
	{
		this.numFilterRows = 1;
		this.include = Include.fromInt(syncFilter.getInclude());
		if (propertyFilters != null) propertyFilters.clear();
		initPropertyFilters(syncFilter.getPropertyFilters());
	}
	
	/**
		Store the current property filters and save them in the give filterData
		map.
	*/
	private void storePropertyFilters(Map<String, PropertyFilterData> filterData)
	{
		for (PropertyFilterPair filterPair : propertyFilters.values())
		{
			PropertyFilter filter = filterPair.primary;
			if (filter != null && !filter.isTemporary())
			{
				// Only save primary filters
				// For filters on filter rows > 0 we must combine <property>[filterrow]
				int rowIndex = filter.getRowIndex();
				filterData.put(filter.getProperty() + (rowIndex > 0 ? "[" + rowIndex + "]" : ""), filter.getData());
			}
		}
	}
	
	/**
		Initialize the property filters by loading them from the given map. Eg.
		revese of {@link #storePropertyFilters(Map)}.
	*/
	private void initPropertyFilters(Map<String, PropertyFilterData> filterData)
	{
		for (Map.Entry<String, PropertyFilterData> entry : filterData.entrySet())
		{
			PropertyFilter pf = PropertyFilter.create(entry.getKey(), entry.getValue());
			if (pf.getRowIndex() >= numFilterRows) 
			{
				numFilterRows = pf.getRowIndex()+1;
			}
			setPropertyFilter(pf);
		}
	}
	
	/**
		Get the type of item this object is storing context for.
	*/
	public Item getItemType()
	{
		return itemType;
	}
	
	/**
		Get the sub-context name of this context.
	*/
	public String getSubContext()
	{
		return subContext;
	}

	/**
		Get the name of this context.
	*/
	public String getName()
	{
		return name;
	}

	/**
		Get the database ID of this context. Can be used to load contexts
		@return The database ID of the context, or 0 if it hasn't been saved to 
			the database
		@since 2.4
	*/
	public int getContextId()
	{
		return contextId;
	}
	
	/**
		Get the ID of the current item in this context.
		@return The ID of the item, or 0 if no item is current
	*/
	public int getId()
	{
		return itemId;
	}
	
	/**
		Set the ID of the current item in this context. This property is stored in 
		the database.
		@param id The ID of the item, or 0 if no item is current
	*/
	public void setId(int id)
	{
		this.itemId = id;
	}
	
	/**
		Get the number of rows the current table page is displaying.
		@return The number of rows
	*/
	public int getRowsPerPage()
	{
		return rowsPerPage;
	}
	
	/**
		Set the number of rows the current table page should show. 
		This property is stored in the database.
		@param rowsPerPage The number of rows to show on a page
	*/
	public void setRowsPerPage(int rowsPerPage)
	{
		this.rowsPerPage = rowsPerPage;
	}
	
	/**
		Get the current table page.
		@return The number of the page, starting at 0
	*/
	public int getPage()
	{
		return page;
	}
	
	/**
		Set the current table page.
		This property is stored in the database.
		@param page The number of the page, starting at 0
	*/
	public void setPage(int page)
	{
		this.page = page;
	}

	/**
		Get the number of filter rows to display on the list page.
		@since 3.5
	*/
	public int getFilterRows()
	{
		return numFilterRows;
	}
	
	/**
		Add filter rows to the table.
		@param numRows The number of rows to add, must be greater than 0
		@return The number of filter rows after adding 
		@since 3.5
	*/
	public int addFilterRows(int numRows)
	{
		if (numRows > 0) numFilterRows += numRows;
		return numFilterRows;
	}
	
	/**
		Remove all filters from the specified filter row. Rows with a higher
		index are shifted down by one step.
		
		@param filterIndex A value between 0 and number of filter rows (-1)
		@return The number of filter rows after removing 
		@since 3.5
	*/
	public int removeFilterRow(int filterIndex)
	{
		if (filterIndex < 0 || filterIndex >= numFilterRows) return numFilterRows;
		
		// Permission filter is on the first row only
		if (filterIndex == 0) permission = null;
		
		// We need to copy remaining filter to a new collection
		if (propertyFilters != null && propertyFilters.size() > 0)
		{
			Map<String, PropertyFilterPair> tmp = new HashMap<String, PropertyFilterPair>();
			for (PropertyFilterPair pair : propertyFilters.values())
			{
				PropertyFilter filter = pair.getEffectiveFilter();
				int rowIndex = filter.getRowIndex();
				if (rowIndex < filterIndex)
				{
					// Copy to new collection as is
					tmp.put(filter.getProperty() + "@" + rowIndex, pair);
				}
				else if (rowIndex > filterIndex)
				{
					rowIndex--;
					// Decrease row index by 1 and copy to new collection
					if (pair.primary != null) pair.primary.setRowIndex(rowIndex);
					if (pair.temporary != null) pair.temporary.setRowIndex(rowIndex);
					tmp.put(filter.getProperty() + "@" + rowIndex, pair);
				}
				// skip filters with rowIndex == filterIndex
			}
			propertyFilters = tmp;
		}
		if (numFilterRows > 1) numFilterRows--;
		return numFilterRows;
	}
	
	/**
		Get the name of the property the current listing is sorted by.
		If multiple properties are used for sorting they are separated by a comma.
		Each property may also be prefixed by a '+' or '-' to override the
		global sort direction.
	*/
	public String getSortProperty()
	{
		return sortProperty;
	}
	
	/**
		Set the name of the property that the current listing is sorted by.
		If you want to sort by multiple properties, separate them with a 
		comma. By default, each property uses the same sort direction ({@link #getSortDirection()},
		but this can be overridden by using '+' or '-' as a prefix for
		a property.
		This property is stored in the database.
		@param sortProperty The name of the property
	*/
	public void setSortProperty(String sortProperty)
	{
		this.sortProperty = sortProperty;
	}

	/**
		If the items are sorted in ascending or descending order.
	*/
	public SortDirection getSortDirection()
	{
		return sortDirection == null ? SortDirection.ASC : sortDirection;
	}
	
	/**
		Set the sort direction (ascending or descending) of the current listing.
		This property is stored in the database.
		@param sortDirection The sort direction
	*/
	public void setSortDirection(SortDirection sortDirection)
	{
		this.sortDirection = sortDirection;
	}

	/**
		Get the current sort property as an <code>Order</code> instance.
		If multiple properties are used for sorting, this method only
		returns the first property.
		@return An Order instance, or null if no sort order has been
			specified
	*/
	public Order getSortOrder()
	{
		Order sortOrder = null;
		String sortProperties = getSortProperty();
		if (sortProperties != null)
		{
			String[] sp = sortProperties.split(",");
			String sortProperty = sp[0];
			String alias = null;
			SortDirection sortDirection = getSortDirection();
			
			if (sortProperty.startsWith("+"))
			{
				sortDirection = SortDirection.ASC;
				sortProperty = sortProperty.substring(1);
			}
			else if (sortProperty.startsWith("-"))
			{
				sortDirection = SortDirection.DESC;
				sortProperty = sortProperty.substring(1);
			}
			if (sortProperty.startsWith("$"))
			{
				// sortProperty is $alias.property or $alias only
				int dotIndex = sortProperty.indexOf(".");
				if (dotIndex == -1)
				{
					// alias only
					alias = sortProperty.substring(1);
					sortProperty = null;
				}
				else
				{
					// alias.property
					alias = sortProperty.substring(1, dotIndex);
					sortProperty = sortProperty.substring(dotIndex+1);
				}
			}
			Expression hqlSortby = Hql.property(alias, sortProperty);
			sortOrder = sortDirection.sortBy(hqlSortby);
		}
		return sortOrder;
	}
	
	/**
		Get a message that should be displayed by a client application as
		soon as it is appropriate to do so. Typically, the 
		{@link #configureQuery(DbControl, EntityQuery, boolean)}
		sets an error message if it couldn't configure the query for some reason. 
	*/
	public String getMessage()
	{
		return message;
	}
	
	/**
		Set a message to be displayed by a client application as soon as
		it is appropriate to do so. Typically, the {@link #configureQuery(DbControl, EntityQuery, boolean)}
		sets an error message if it couldn't configure the query for some reason. 
		This property is not stored in the database.
		@param message The message
	*/
	public void setMessage(String message)
	{
		this.message = message;
	}
	
	/**
		Get a <code>Set</code> object to specify {@link Include} options for 
		a query. This property is stored in the database.
		@see EntityQuery
	*/
	public Set<Include> getInclude()
	{
		if (include == null) include = EnumSet.noneOf(Include.class);
		return include;
	}
	
	/**
	 	Specify the permission the logged in user must have for items returned
	 	by a query. This property is not stored in the database.
		@see EntityQuery#setItemPermission(Permission)
	*/
	public void setItemPermission(Permission permission)
	{
		this.permission = permission;
	}
	
	/**
	 	The permission the logged in user must have for items returned
	 	by a query. This property is not stored in the database.
		@see EntityQuery#getItemPermission()
	*/
	public Permission getItemPermission()
	{
		return permission;
	}
	
	/**
		Get a <code>Set</code> object to specify the ID:s of all items that are
		currently selected. This property is not stored in the database.
	*/
	public Set<Integer> getSelected()
	{
		if (selected == null) selected = new HashSet<Integer>();
		return selected;
	}
	
	/**
		Set a property filter. If a filter for the same property already exists it
		is replaced by the new filter. Note that each property may hold two
		filters, one primary and one temporary. Only primary filters are stored
		in the database.
		@param filter The new property filter
		@see PropertyFilter#isTemporary()
	*/
	public void setPropertyFilter(PropertyFilter filter)
	{
		if (propertyFilters == null)
		{
			propertyFilters = new HashMap<String, PropertyFilterPair>();
		}
		String filterKey = filter.getProperty() + "@" + filter.getRowIndex();
		PropertyFilterPair filterPair = propertyFilters.get(filterKey);
		if (filterPair == null)
		{
			filterPair = new PropertyFilterPair();
			propertyFilters.put(filterKey, filterPair);
		}
		if (filter.isTemporary())
		{
			filterPair.temporary = filter;
		}
		else
		{
			filterPair.primary = filter;
		}
	}
	
	/**
		Get a filter for a property on filter row 0. 
		@see #getPropertyFilter(String, int)
	*/
	public PropertyFilter getPropertyFilter(String property)
	{
		return getPropertyFilter(property, 0);
	}
	
	/**
		Get a filter for a property. If the given property has a temporary
		filter, it is returned, otherwise it's primary filter is returned.
		@param property The name of the property
		@param filterRow The index of the filter row should be between 0 and number of filter rows (-1)
		@return A <code>PropertyFilter</code> object or
			null if no filter exists
		@since 3.5
	*/
	public PropertyFilter getPropertyFilter(String property, int filterRow)
	{
		PropertyFilter filter = null;
		if (propertyFilters != null)
		{
			PropertyFilterPair filterPair = propertyFilters.get(property + "@" + filterRow);
			if (filterPair != null) 
			{
				filter = filterPair.getEffectiveFilter();
			}
		}
		return filter;
	}
	
	/**
		Get a collection of all property filters stored in this context.
		The returned collection may contain both primary and temporary
		filters, but only one filter for each property.
		@return A Collection containsing PropertyFilter objects or null
	*/
	public Collection<PropertyFilter> getPropertyFilters()
	{
		if (propertyFilters == null) return null;
		List<PropertyFilter> filters = new ArrayList<PropertyFilter>(propertyFilters.size());
		for (PropertyFilterPair filterPair : propertyFilters.values())
		{
			filters.add(filterPair.getEffectiveFilter());
		}
		return filters;
	}
	
	/**
		Get a collection of all property filters stored in this context for 
		a given filter row.
		The returned collection may contain both primary and temporary
		filters, but only one filter for each property.
		@param filterRow The index of the filter row should be between 0 and number of filter rows (-1)
		@return A Collection containsing PropertyFilter objects or null
		@since 3.5
	*/
	public Collection<PropertyFilter> getPropertyFilters(int filterRow)
	{
		if (propertyFilters == null) return null;
		List<PropertyFilter> filters = new ArrayList<PropertyFilter>(propertyFilters.size());
		for (PropertyFilterPair filterPair : propertyFilters.values())
		{
			PropertyFilter filter = filterPair.getEffectiveFilter();
			if (filter.getRowIndex() == filterRow) 
			{
				filters.add(filter);
			}
		}
		return filters;
	}

	
	/**
		Get the total number of property filters stored in this
		context.
		@since 2.12
	*/
	public int getNumPropertyFilters()
	{
		return (propertyFilters == null ? 0 : propertyFilters.size()) + (permission != null ? 1 : 0);
	}
	
	/**
		Get the number of property filters stored in this
		context for the given filter row.
		@param filterRow The index of the filter row should be between 0 and number of filter rows (-1)
		@since 3.5
	*/
	public int getNumPropertyFilters(int filterRow)
	{
		if (propertyFilters == null) return 0;
		int count = filterRow == 0 && permission != null ? 1 : 0;
		for (PropertyFilterPair filterPair : propertyFilters.values())
		{
			PropertyFilter filter = filterPair.getEffectiveFilter();
			if (filter.getRowIndex() == filterRow) count++;
		}
		return count;
	}
	
	/**
		Checks if there is a filter for the given property regardless of
		filter row. Shortcut for checking each filter row at a time and
		{@link #getPropertyFilter(String, int)} for non-null.
		@param property The name of the property
		@return TRUE if at least one filter exists, FALSE if not
		@since 3.5
	*/
	public boolean hasPropertyFilter(String property)
	{
		if (propertyFilters == null) return false;
		for (PropertyFilterPair filterPair : propertyFilters.values())
		{
			if (property.equals(filterPair.getEffectiveFilter().getProperty())) return true;
		}
		return false;
	}
	
	/**
		Get the value of a property filter on row 0. 
		@see #getPropertyValue(String, int)
	*/
	public String getPropertyValue(String property)
	{
		return getPropertyValue(property, 0);
	}
	
	/**
		Get the value of a property filter on the given filter row. 
		It is a shortcut for <code>getPropertyFilter(property, filterRow).getValue()</code>
		but also handles the case of a non-existing filter for the
		specfied property.
		@param property The name of the property
		@param filterRow The index of the filter row should be between 0 and number of filter rows (-1)
		@return The value as a string
		@since 3.5
	*/
	public String getPropertyValue(String property, int filterRow)
	{
		String value = null;
		PropertyFilter filter = getPropertyFilter(property, filterRow);
		if (filter != null)
		{
			value = filter.getValue();
		}
		return value;
	}
	
	/**
		Remove a property filter on filter row 0.
		@see #removePropertyFilter(String, int)
	*/
	public void removePropertyFilter(String property)
	{
		removePropertyFilter(property, 0);
	}
	
	/**
		Remove a property filter on the given filter row. This method will remove both
		a primary and a temporary filter.
		@param property The name of the property
		@param filterRow The index of the filter row should be between 0 and number of filter rows (-1)
		@since 3.5
	*/
	public void removePropertyFilter(String property, int filterRow)
	{
		if (propertyFilters != null)
		{
			propertyFilters.remove(property + "@" + filterRow);
		}
	}

	/**
		Remote a temporary filter for the given property on filter row 0.
		@see #removeTemporaryFilter(String, int)
		@since 2.16
	*/
	public void removeTemporaryFilter(String property)
	{
		removeTemporaryFilter(property, 0);
	}
	
	/**
		Remote a temporary filter for the given property. If
		a primary filter exists it will now be the effective filter.
		
		@param property The name of the property
		@param filterRow The index of the filter row should be between 0 and number of filter rows (-1)
		@since 3.5
	*/
	public void removeTemporaryFilter(String property, int filterRow)
	{
		if (propertyFilters == null) return;
		PropertyFilterPair filterPair = propertyFilters.get(property + "@" + filterRow);
		if (filterPair != null)
		{
			filterPair.temporary = null;
			if (filterPair.primary == null) propertyFilters.remove(property);
		}
	}

	
	public Object getPropertyObject(String property)
	{
		Object value = null;
		PropertyFilter filter = getPropertyFilter(property);
		if (filter != null)
		{
			value = filter.getValueAsObject();
		}
		return value;
	}
	
	/**
		Remove all property filter from this context.
	*/
	public void removeAllPropertyFilters()
	{
		if (propertyFilters != null)
		{
			propertyFilters.clear();
		}
		numFilterRows = 1;
	}
	
	/**
		Remove all temporary filters.
		@since 2.16
	*/
	public void removeAllTemporaryFilters()
	{
		if (propertyFilters == null) return;
		Iterator<PropertyFilterPair> it = propertyFilters.values().iterator();
		while (it.hasNext())
		{
			PropertyFilterPair filterPair = it.next();
			filterPair.temporary = null;
			if (filterPair.primary == null)
			{
				it.remove();
			}
		}
	}
	
	/**
		Create a restriction from a property filter on filter row 0.
		@see #getFilterRestriction(String, int, DbControl, EntityQuery)
		@since 2.15
	*/
	public Restriction getFilterRestriction(String property, DbControl dc, EntityQuery query)
	{
		return getFilterRestriction(property, 0, dc, query);
	}
	
	/**
		Create a restriction from a property filter on the given filter row.
		@param property The property
		@param filterRow The index of the filter row should be between 0 and number of filter rows (-1)
		@param dc A DbControl that should be used if database access is needed
		@param query The query that the filter is going to be used in (can be
			null for most properties)
		@return A restriction, or null if no filter is active for the given
			property
		@since 3.5
	*/
	public Restriction getFilterRestriction(String property, int filterRow, DbControl dc, EntityQuery query)
	{
		Restriction r = null;
		PropertyFilter filter = getPropertyFilter(property, filterRow);
		if (filter != null)
		{
			r = filter.getRestriction(dc, query);
		}
		return r;
	}
	
	/**
		Add an item to the "recently used items" list. Each type of item
		has their own list which may contain at most <code>maxInList</code>
		items. 
		<p>
		Note! The lists are stored as string of colon-separated ID:s in
		the setting given by the item type. For example the recently used
		protocols are found in the setting: 
		<code>getSetting("PROTOCOL.recent")</code>. Client code is advised to
		not change these settings.
		
		@param item The item to add
		@param maxInList The maximum number of items in the list
	 */
	public void setRecent(BasicItem item, int maxInList)
	{
		setRecent(item.getType().name(), Integer.toString(item.getId()), maxInList);
	}
	
	/**
		Add an item to the "recently used items" list when used for a specific
		item subtype. Each type of item+subtype has their own list which may contain at 
		most <code>maxInList</code> items. 
		<p>
		Note! The lists are stored as string of colon-separated ID:s in
		the setting given by the item type. For example the recently used
		protocols are found in the setting: 
		<code>getSetting("PROTOCOL.#subtype-id.recent")</code>. Client code is advised to
		not change these settings.
		
		@param item The item to add
		@param subtype The subtype of the parent item (can be null)
		@param maxInList The maximum number of items in the list
		@since 3.0
	 */
	public void setRecent(BasicItem item, ItemSubtype subtype, int maxInList)
	{
		if (subtype == null)
		{
			setRecent(item, maxInList);
		}
		else
		{
			setRecent(item, "#"+subtype.getId(), maxInList);
		}
	}
	
	/**
		Add an item to the "recently used items" list. Each type of item
		has their own list and sublists which may contain at 
		most <code>maxInList</code> items. 
		<p>
		Note! The lists are stored as string of colon-separated ID:s in
		the setting given by the item type and sublist name. For example the 
		recently used protocols are found in the setting: 
		<code>getSetting("PROTOCOL.subList.recent")</code>. Client code is advised to
		not change these settings.

		@param item The item to add
		@param subList The sublist
		@param maxInList The maximum number of items in the list
		@since 2.5
	*/
	public void setRecent(BasicItem item, String subList, int maxInList)
	{
		setRecent(item.getType().name() + "." + subList, Integer.toString(item.getId()), maxInList);
	}
	
	/**
		Add a value to the list of recently used values for the specified key.
		Each key has it's own list of values which may contain at most 
		<code>maxInList</code> values. If the value is already in the list
		it is moved to the front of the list.
		<p>
		Note! The lists are stored as string of colon-separated ID:s in
		the setting given by the key. For example, if key is "RAWDATATYPE",
		the recently used values are found in the setting: 
		<code>getSetting("RAWDATATYPE.recent")</code>. Client code is advised to
		not change these settings.

		@param key The key to the list
		@param value The value to add
		@param maxInList The maximum number of values in the list
	*/
	public void setRecent(String key, String value, int maxInList)
	{
		List<String> recent = getRecent(key);
		int existingIndex = recent.indexOf(value);
		if (existingIndex >= 0) recent.remove(existingIndex);
		recent.add(0, value);
		if (recent.size() > maxInList) recent = recent.subList(0, maxInList);
		setRecent(key, recent);
	}
	
	/**
		Get the number of recently used items in the list.
		@param itemType The type of items
	*/
	public int getNumRecent(Item itemType)
	{
		return getRecent(itemType.name()).size();
	}
	
	/**
		Get the number of recently used items in the list.
		@param itemType The type of items
		@param subtype The subtype of the parent item (can be null)
		@since 3.0
	*/
	public int getNumRecent(Item itemType, ItemSubtype subtype)
	{
		if (subtype == null)
		{
			return getNumRecent(itemType);
		}
		else
		{
			return getNumRecent(itemType, "#"+subtype.getId());
		}
	}
	
	/**
		Get the number of recently used items in a list with a sublist.
		@param itemType The type of items
		@param subList The name of the sublist
		@since 2.5
	*/
	public int getNumRecent(Item itemType, String subList)
	{
		return getRecent(itemType.name() + "." + subList).size();
	}
	
	/**
		Get the number of recently used value in the list.
		@param key The key to the list
	*/
	public int getNumRecent(String key)
	{
		return getRecent(key).size();
	}
	
	/**
		Get all recently used items of the specified type. The list
		only contains items that the logged in user has read permission
		to and that hasn't been deleted. Thus, the {@link #getNumRecent(Item)}
		may return a larger number than what is actually returned in this
		list.
		
		@param dc The DbControl used to access the database
		@param itemType The type of items
		@return A list with the items
	*/
	public List<? extends BasicItem> getRecent(DbControl dc, Item itemType)
	{
		return loadRecent(dc, itemType, itemType.name());
	}
	
	/**
		Get all recently used items of the specified type. The list
		only contains items that the logged in user has read permission
		to and that hasn't been deleted. Thus, the {@link #getNumRecent(Item, ItemSubtype)}
		may return a larger number than what is actually returned in this
		list.
		
		@param dc The DbControl used to access the database
		@param itemType The type of items
		@param subtype The subtype of the parent item (can be null)
		@return A list with the items
		@since 3.0
	*/
	public List<? extends BasicItem> getRecent(DbControl dc, Item itemType, ItemSubtype subtype)
	{
		if (subtype == null)
		{
			return getRecent(dc, itemType);
		}
		else
		{
			return getRecent(dc, itemType, "#" + subtype.getId());
		}
	}
	
	/**
		Get all recently used items of the specified type in a given sublist.
		The list only contains items that the logged in user has read permission
		to and that hasn't been deleted. Thus, the {@link #getNumRecent(Item)}
		may return a larger number than what is actually returned in this
		list.
		
		@param dc The DbControl used to access the database
		@param itemType The type of items
		@param subList The name of the sublist
		@return A list with the items
		@since 2.5
	*/
	public List<? extends BasicItem> getRecent(DbControl dc, Item itemType, String subList)
	{
		return loadRecent(dc, itemType, itemType.name() + "." + subList);
	}
	
	private List<? extends BasicItem> loadRecent(DbControl dc, Item itemType, String key)
	{
		List<String> recentIds = getRecent(key);
		List<BasicItem> recent = new ArrayList<BasicItem>(recentIds.size());
		for (String id : recentIds)
		{
			try
			{
				recent.add(itemType.getById(dc, Integer.parseInt(id)));
			}
			catch (Throwable t)
			{
				Application.getLogger().debug("Failed loading recent " + 
					itemType.name()+" [id="+id+"]", t);
			}
		}
		return recent;
	}
	
	
	public BasicItem getRecent(DbControl dc, Item itemType, int index)
	{
		return loadRecent(dc, itemType, itemType.name(), index);
	}
	
	public BasicItem getRecent(DbControl dc, Item itemType, String subList, int index)
	{
		return loadRecent(dc, itemType, itemType.name() + "." + subList, index);
	}
	
	private BasicItem loadRecent(DbControl dc, Item itemType, String key, int index)
	{
		List<String> recentIds = getRecent(key);
		BasicItem item = null;
		if (index >= 0 && index < recentIds.size())
		{
			String id = recentIds.get(index);
			try
			{
				item = itemType.getById(dc, Integer.parseInt(id));
			}
			catch (Throwable t)
			{
				Application.getLogger().debug("Failed loading recent " + 
					itemType.name()+" [id="+id+"]", t);
			}
			
		}
		return item;
	}

	
	/**
		Get all recently used values for the specified key. Changes 
		to the list are not recorded in the database. Use {@link 
		#setRecent(String, String, int)} instead.
		@param key The key to the list
		@return A list containing the values or an empty list if
			no recently used value has been stored
	*/
	public List<String> getRecent(String key)
	{
		String recent = getSetting(getRecentSettingName(key));
		if (recent == null)
		{
			return new ArrayList<String>();
		}
		else
		{
			return EncodeUtil.decodeAll(Arrays.asList(recent.split("\n")), new TabCrLfEncoderDecoder(false));
		}
	}
	
	public String getRecent(String key, int index)
	{
		List<String> recent = getRecent(key);
		return index >= 0 && index < recent.size() ? recent.get(index) : null;
	}
	
	/**
		Clear the recently used items list from items of the specified type.
		@param itemType The type of items
	*/
	public void clearRecent(Item itemType)
	{
		setRecent(itemType.name(), null);
	}

	/**
		Clear the recently used items list from items of the specified type.
		@param itemType The type of items
		@param subtype The subtype of the parent item (can be null)
		@since 3.0
	*/
	public void clearRecent(Item itemType, ItemSubtype subtype)
	{
		if (subtype == null)
		{
			clearRecent(itemType);
		}
		else
		{
			clearRecent(itemType, "#" + subtype.getId());
		}
	}
	

	/**
		Clear the recently used items sublist from items of the specified type.
		@param itemType The type of items
		@param subList The name of the sublist
	*/
	public void clearRecent(Item itemType, String subList)
	{
		setRecent(itemType.name() + "." + subList, null);
	}
	
	/**
		Clear the recently used values list.
		@param key The key to the list
	*/
	public void clearRecent(String key)
	{
		setRecent(key, null);
	}
	
	/**
		Clear all recently used values lists.
	*/
	public void clearAllRecent()
	{
		if (settings != null)
		{
			Iterator<Map.Entry<String, String>> it = settings.entrySet().iterator();
			while (it.hasNext())
			{
				String key = it.next().getKey();
				if (key.endsWith(".recent"))
				{
					it.remove();
				}
			}
		}
	}
	
	private String getRecentSettingName(String key)
	{
		return key+".recent";
	}

	private void setRecent(String key, List<String> recent)
	{
		String setting = null;
		if (recent != null && recent.size() > 0)
		{
			setting = EncodeUtil.encodeAndJoin(recent, new TabCrLfEncoderDecoder(false), "\n", true);
		}
		setSetting(getRecentSettingName(key), setting);
	}
	
	/**
		Get a value.
		@param name The name of the value
		@return The current value or null if no value exists
		@see #setSetting(String, String)
	*/
	public String getSetting(String name)
	{
		return settings == null ? null : settings.get(name);
	}
	
	/**
		Set a value. This property is stored in the database.
		@param name The name of the value
		@param value The value
		@return The old value or null if no old value existed.
		@see #getSetting(String)
	*/
	public String setSetting(String name, String value)
	{
		if (settings == null) settings = new HashMap<String, String>();
		return settings.put(name, value);
	}
	
	/**
		Remove a value.
		@param name The name of the value
		@return The old value or null if no old value existed.
	*/
	public String removeSetting(String name)
	{
		return settings == null ? null : settings.remove(name);
	}
	
	/**
		Get an object.
		@param name The name of the value
		@return The current value or null if no value exists
		@see #setObject(String, Object)
	*/
	public Object getObject(String name)
	{
		return objects == null ? null : objects.get(name);
	}
	
	/**
		Set a value. This property is not stored in the database.
		@param name The name of the value
		@param value The value
		@return The old value or null if no old value existed.
		@see #getObject(String)
	*/
	public Object setObject(String name, Object value)
	{
		if (objects == null) objects = new HashMap<String, Object>();
		return objects.put(name, value);
	}
	
	/**
		Remove a value.
		@param name The name of the value
		@return The old value or null if no old value existed.
	*/
	public Object removeObject(String name)
	{
		return objects == null ? null : objects.remove(name);
	}
	
	/**
		Copies all objects and temporary filters from this context
		to another context.
		@since 3.6
	*/
	public void copyObjectsAndTemporaryFilters(ItemContext other)
	{
		if (objects != null && objects.size() > 0)
		{
			if (other.objects == null) other.objects = new HashMap<String, Object>();
			other.objects.putAll(objects);
		}
		if (propertyFilters != null && propertyFilters.size() > 0)
		{
			for (PropertyFilterPair filter : propertyFilters.values())
			{
				if (filter.temporary != null)
				{
					other.setPropertyFilter(filter.temporary);
				}
			}
		}
	}

	public Query getQuery()
	{
		return query;
	}
	
	public void setQuery(Query query)
	{
		this.query = query;
	}
	
	/**
		Set a filter for this context that is used to inspect which properties
		that are allowed to be used in queries. If a filter has been set the
		{@link #configureQuery(DbControl, EntityQuery, boolean)} and
		{@link #configureQuery(DbControl, SqlQuery, List)} methods will
		check each property before it is used for sorting, filtering, etc.
		
		@param filter A filter or null to disable property inspection
		@since 3.1
	*/
	public void setPropertyInspector(Filter<String> filter)
	{
		this.propertyInspector = filter;
	}
	
	// Do not auto-join queries containing filter properties with the following prefixes
	private static final Set<String> NO_AUTO_JOIN_PREFIXES = new HashSet<String>(
		Arrays.asList("£", "&", "!", "¤", "|"));
	
	/**
		Use the settings in this context to configure a query. 
		<ul>
		<li>{@link EntityQuery#setReturnTotalCount(boolean)} is set to true.
		<li>{@link EntityQuery#setFirstResult(int)} and {@link EntityQuery#setMaxResults(int)} 
			is set using {@link #getPage()} and {@link #getRowsPerPage()} settings.
			
		<li>The {@link #getSortProperty()} and {@link #getSortDirection()} is
			used to sort the query. If the sort property starts with a dollar ($)
			sign it is interpreted as a <code>alias.property</code> value, ignoring
			the root alias of the query. The sort property may contain multiple properties
			if they are separated by a comma. Each property may also be prefixed by
			a '+' (ASC) or '-' (DESC) to override the global sort direction.
			
		<li>{@link #getInclude()} is used to set {@link EntityQuery#include(Include[])}
			or @link EntityQuery#exclude(Include[])} options.
			
		<li>All property filters are added as {@link Restriction}:s to the 
			query.
			
		</ul>
		
		@param dc The DbControl that will be used to execute the query
		@param query The query to configure
		@param autoLeftJoin If true, and the sort property and filters are checked
			for associations to other items, which are automatically left joined to
			the query (instead of the hibernate default of inner join) (this feature
			is experimental)
	 	@throws BaseException If configuring the query fails.
	 	@since 2.8
	*/
	public void configureQuery(DbControl dc, EntityQuery query, boolean autoLeftJoin)
		throws BaseException
	{
		// Add page information
		if (!noPaging)
		{
			query.setReturnTotalCount(true);
			query.setFirstResult(getPage() * getRowsPerPage());
			query.setMaxResults(getRowsPerPage());
		}

		// Temporary storage for all associations that must be left joined
		Set<String> leftJoins = autoLeftJoin ? new TreeSet<String>(new StringLengthComparator()) : null;
		Set<String> sortedProperties = autoLeftJoin ? new HashSet<String>() : null;
		
		// Add sort order information
		String sortProperties = getSortProperty();
		boolean isSortingOnIdProperty = false;
		if (sortProperties != null)
		{
			String[] sp = sortProperties.split(",");
			for (String sortProperty : sp)
			{
				if (propertyInspector == null || propertyInspector.evaluate(sortProperty))
				{
					SortDirection sortDirection = getSortDirection();
					
					Expression hqlSortby = null;
					String alias = null;
					
					// Check if sort direction has been overridden
					if (sortProperty.startsWith("-"))
					{
						sortDirection = SortDirection.DESC;
						sortProperty = sortProperty.substring(1);
					}
					else if (sortProperty.startsWith("+"))
					{
						sortDirection = SortDirection.ASC;
						sortProperty = sortProperty.substring(1);
					}
					
					// Extract the 'alias' part if there is one
					if (sortProperty.startsWith("$"))
					{
						// sortProperty is $alias.property or $alias only
						int dotIndex = sortProperty.indexOf(".");
						if (dotIndex == -1)
						{
							// alias only
							alias = sortProperty.substring(1);
							sortProperty = null;
						}
						else
						{
							// alias.property
							alias = sortProperty.substring(1, dotIndex);
							sortProperty = sortProperty.substring(dotIndex+1);
						}
					}
					if (sortProperty.startsWith("#"))
					{
						// Allow sorting on annotation types with multiplicity = 1
						int annotationTypeId = (Integer)Type.INT.parseString(sortProperty.substring(1));
						AnnotationType at = AnnotationType.getById(dc, annotationTypeId);
						if (at.getMultiplicity() == 1) hqlSortby = Annotations.expression(alias, at);
					}
					else
					{
						// left join if sort property is association 
						if (autoLeftJoin && sortProperty != null)
						{
							int dotIndex = sortProperty.lastIndexOf(".");
							if (dotIndex >= 0)
							{
								String toJoin = sortProperty.substring(0, dotIndex);
								if (alias != null) toJoin = "$" + alias + "." + toJoin;
								leftJoins.add(toJoin);
								sortedProperties.add(toJoin);
							}
						}
						isSortingOnIdProperty |= "id".equals(sortProperty);
						hqlSortby = Hql.property(alias, sortProperty);
					}
					if (hqlSortby != null)
					{
						query.order(sortDirection.sortBy(hqlSortby));
					}
				}
			}
		}
		// To make sure we get a reproducable ordering when there is no
		// other ordering or of the ordering is made on non-unique columns
		if (!isSortingOnIdProperty && !noPaging)
		{
			query.order(getSortDirection().sortBy(Hql.property("id")));
		}
		
		// Add include/exclude information
		Set<Include> includes = getInclude();
		for (Include i : Include.values())
		{
			if (includes.contains(i)) 
			{
				query.include(i);
			}
			else
			{
				query.exclude(i);
			}
		}
		
		// Permission info
		if (permission != null) query.setItemPermission(permission);
		
		// Add property filters
		if (propertyFilters != null)
		{
			Map<Integer, List<Restriction>> allRestrictions = new HashMap<Integer, List<Restriction>>();
			for (PropertyFilterPair filterPair : propertyFilters.values())
			{
				PropertyFilter filter = filterPair.getEffectiveFilter();
				if (propertyInspector == null || propertyInspector.evaluate(filter.getProperty()))
				{
					try
					{
						Restriction r = filter.getRestriction(dc, query);
						if (r != null)
						{
							int rowIndex = filter.getRowIndex();
							List<Restriction> rowRestrictions = allRestrictions.get(rowIndex);
							if (rowRestrictions == null)
							{
								rowRestrictions = new ArrayList<Restriction>();
								allRestrictions.put(rowIndex, rowRestrictions);
							}
							rowRestrictions.add(r);
							// left join if filter property is association 
							if (autoLeftJoin)
							{
								String filterProperty = filter.getProperty();
								if (filterProperty != null && !NO_AUTO_JOIN_PREFIXES.contains(filterProperty.substring(0,  1)))
								{
									int lastDotIndex = filterProperty.lastIndexOf('.');
									int firstDotIndex = filterProperty.indexOf('.');
									if (lastDotIndex >= 0)
									{
										filterProperty = filterProperty.substring(0, lastDotIndex);
										if (!filterProperty.startsWith("$") || lastDotIndex != firstDotIndex)
										{
											leftJoins.add(filterProperty);
										}
									}
								}
							}
						}
					}
					catch (Throwable ex)
					{
						String msg = "Could not filter on '" + getItemType() + "." + filter.getProperty() + "'";
						Application.getLogger().warn(msg, ex);
						setMessage(msg + ": " + ex.getMessage());
					}
				}
			}
			if (allRestrictions.size() > 0)
			{
				Restriction[] rows = new Restriction[allRestrictions.size()];
				int rowNo = 0;
				for (List<Restriction> rowRestrictions : allRestrictions.values())
				{
					rows[rowNo] = Restrictions.nullSafeAnd(rowRestrictions);
					rowNo++;
				}
				query.restrict(Restrictions.or(rows));
			}
		}
		
		// Add left joins
		if (autoLeftJoin)
		{
			int i = 0;
			// Use fetch joins for columns appearing in the ORDER BY clause
			// if the query is distinct and the database requires it
			boolean selectOrderBy = query.isDistinct() && 
				HibernateUtil.getDbEngine().selectOrderByColumnsIfDistinct();
			for (String property : leftJoins)
			{
				boolean fetch = selectOrderBy && sortedProperties.contains(property);
				String alias = null;
				if (property.startsWith("@") || property.startsWith("£") || property.startsWith("&") || property.startsWith("§")) 
				{
					property = property.substring(1);
				}
				if (property.startsWith("$"))
				{
					int dotIndex = property.indexOf('.');
					alias = property.substring(1, dotIndex);
					property = property.substring(dotIndex+1);
				}
				i++;
				query.join(Hql.leftJoin(alias, property, "alj"+i, null, fetch));
			}
		}
	}
	
	/**
		Copy settings from another context to this context. Only settings which doesn't 
		have been set in this context is copied from the other context, unless
		the <code>forceAll</code> parameter is set.
	*/
	void copyFrom(ItemContext other, boolean forceAll)
	{
		if (this.rowsPerPage == -1 || forceAll) 
		{
			this.rowsPerPage = other.rowsPerPage;
		}
		if (this.page == -1 || forceAll) 
		{
			this.page = other.page;
		}
		if (this.sortProperty == null || forceAll) 
		{
			this.sortProperty = other.sortProperty;
		}
		if (this.sortDirection == null || forceAll) 
		{
			this.sortDirection = other.sortDirection;
		}
		if ((this.include == null  || forceAll) && other.include != null) 
		{
			this.getInclude().addAll(other.include);
		}
		if ((this.settings == null || forceAll) && other.settings != null) 
		{
			this.settings = new HashMap<String, String>(other.settings);
		}
		if ((this.objects == null || forceAll) && other.objects != null) 
		{
			this.objects = new HashMap<String, Object>(other.objects);
		}
	}
	
	/**
		Add a filter to the given query that restricts it to only return items that
		have been selected. If no items have been selected the query is not modified.
		@param query The query to filter
		@since 2.17
	*/
	public void filterOnSelectedItems(EntityQuery query)
	{
		Set<Integer> selected = getSelected();
		if (selected.isEmpty()) return;
		query.restrict(Restrictions.in(Hql.property("id"), Expressions.parameter("selectedItems", selected, Type.INT)));
	}
	
	/**
		@deprecated In 3.1, use {@link #configureQuery(DbControl, SqlQuery, List)} instead
	*/
	@Deprecated
	public void configureQuery(DbControl dc, AbstractSqlQuery query, List<String> selectionList)
	{
		configureQuery(dc, (SqlQuery)query, selectionList);
	}

	/**
		Use the settings in this context to configure a dynamic query. 
		<ul>
		<li>{@link DynamicQuery#setReturnTotalCount(boolean)} is set to true.
		<li>{@link DynamicQuery#setFirstResult(int)} and {@link DynamicQuery#setMaxResults(int)} 
			is set using {@link #getPage()} and {@link #getRowsPerPage()} settings.
		</ul>
		
		@param dc A DbControl to use if the database needs to be accessed
		@param query The query to configure
		@param selectionList A list of expression that we want to select 
			from the query
	 	@throws BaseException If the configuring the query fails.
	 	@since 3.1
	*/
	public void configureQuery(DbControl dc, SqlQuery query, List<String> selectionList)
		throws BaseException
	{
	
		// Add page information
		query.setReturnTotalCount(true);
		query.setFirstResult(getPage() * getRowsPerPage());
		query.setMaxResults(getRowsPerPage());
		

		// Add selection elements
		if (selectionList != null)
		{
			for (String selectProperty : selectionList)
			{
				if (propertyInspector == null || propertyInspector.evaluate(selectProperty))
				{
					Select select = getDynamicSelect(dc, query, selectProperty); //, joins);
					if (select != null)
					{
						query.select(select);
					}
				}
			}
		}
		
		// Add sort order information
		String sortProperties = getSortProperty();
		if (sortProperties != null)
		{
			boolean selectSortProperties = 
				query.isDistinct() && HibernateUtil.getDbEngine().selectOrderByColumnsIfDistinct();
			int sortIndex = 0;
			String[] sp = sortProperties.split(",");
			for (String sortProperty : sp)
			{
				if (propertyInspector == null || propertyInspector.evaluate(sortProperty))
				{
					SortDirection sortDirection = getSortDirection();
					
					// Check if sort direction has been overridden
					if (sortProperty.startsWith("-"))
					{
						sortDirection = SortDirection.DESC;
						sortProperty = sortProperty.substring(1);
					}
					else if (sortProperty.startsWith("+"))
					{
						sortDirection = SortDirection.ASC;
						sortProperty = sortProperty.substring(1);
					}

					Expression sortby = getDynamicExpression(dc, query, sortProperty);
					if (sortby != null)
					{
						query.order(sortDirection.sortBy(sortby));
						if (selectionList == null || !selectionList.contains(sortProperty))
						{
							// We must add the sort property to the selection list
							query.select(Selects.expression(sortby, "sorted" + sortIndex));
						}
					}
					sortIndex++;
				}
			}
		}

		// Add property filters
		if (propertyFilters != null)
		{
			Map<Integer, List<Restriction>> allRestrictions = new HashMap<Integer, List<Restriction>>();
			for (PropertyFilterPair filterPair : propertyFilters.values())
			{
				PropertyFilter filter = filterPair.getEffectiveFilter();
				if (propertyInspector == null || propertyInspector.evaluate(filter.getProperty()))
				{
					try
					{
						Restriction r = filter.getDynamicRestriction(dc, query);
						if (r != null)
						{
							int rowIndex = filter.getRowIndex();
							List<Restriction> rowRestrictions = allRestrictions.get(rowIndex);
							if (rowRestrictions == null)
							{
								rowRestrictions = new ArrayList<Restriction>();
								allRestrictions.put(rowIndex, rowRestrictions);
							}
							rowRestrictions.add(r);

						}
					}
					catch (Throwable ex)
					{
						setMessage(ex.getClass().getSimpleName() + ": " + ex.getMessage());
						ex.printStackTrace();
					}
				}
			}
			if (allRestrictions.size() > 0)
			{
				Restriction[] rows = new Restriction[allRestrictions.size()];
				int rowNo = 0;
				for (List<Restriction> rowRestrictions : allRestrictions.values())
				{
					rows[rowNo] = Restrictions.nullSafeAnd(rowRestrictions);
					rowNo++;
				}
				query.restrict(Restrictions.or(rows));
			}
		}
	}

	/**
		@deprecated In 3.9.2, use {@link #getDynamicExpression(DbControl, SqlQuery, String)} instead
	*/
	@Deprecated
	public static Expression getDynamicExpression(DbControl dc, String propertyDef)
	{
		return getDynamicExpression(dc, null, propertyDef);
	}

	
	/**
		Create an {@link Expression} from a string. If the string starts with:
		
		<ul>
		<li>%: The rest of the string is the channel number. We use the 
			{@link Dynamic#column(VirtualColumn)} to create the expression
			for {@link VirtualColumn#channelRaw(int)}

		<li>$: The rest of the string is a raw data property name. We use
			the {@link Dynamic#rawData(String)} to create the expression.

		<li>@@: The rest of the string is a reporter property name. We use
			the {@link Dynamic#reporter(String, Boolean)} with to create the 
			expression. The second parameter is FALSE to force using the master
			reporter table.

		<li>@: The rest of the string is a reporter property name. We use
			the {@link Dynamic#reporter(String)} to create the expression
			and let the query decide if it should use the cloned or master 
			reporter table.
			
		<li>#: The rest of the string is the id of an extra value. 
			We use the {@link Dynamic#extraValue(ExtraValue)} to create the expression and
			{@link DynamicSpotQuery#joinExtraValue(ExtraValue, JoinType)} to automatically
			join the extra values.
			
		<li>=: The rest of the string is JEP expression. It is converted
			to an Expression by the {@link Jep#formulaToExpression(String, JepFunction[])}
			method. 
		</ul>
		
		It also recognises the following strings for columns to use by
		the {@link Dynamic#column(VirtualColumn)} method:
		<ul>
		<li>POSITION: {@link VirtualColumn#POSITION}
		<li>COLUMN: {@link VirtualColumn#COLUMN}
		</ul>

		@param dc A DbControl to use if the database needs to be accessed
		@param propertyDef The string to use for creating the expression.
		@return An expression object
		@since 3.9.2
	*/
	public static Expression getDynamicExpression(DbControl dc, SqlQuery query, String propertyDef)
	{
		Expression e = null;
		if (propertyDef.startsWith("%"))
		{
			// propertyDef is %channel_number
			int channel = Integer.parseInt(propertyDef.substring(1));
			e = Dynamic.column(VirtualColumn.channelRaw(channel));
		}
		else if (propertyDef.startsWith("$"))
		{
			// propertyDef is $rawDataPropertyName
			e = Dynamic.rawData(propertyDef.substring(1));
		}
		else if (propertyDef.startsWith("@@"))
		{
			// propertyDef is @@reporterPropertyName (from master table)
			e = Dynamic.reporter(propertyDef.substring(2), false);
		}
		else if (propertyDef.startsWith("@"))
		{
			// propertyDef is @reporterPropertyName
			e = Dynamic.reporter(propertyDef.substring(1));
		}
		else if (propertyDef.startsWith("#"))
		{
			// propertyDef id of extra value: #id
			int extraId = net.sf.basedb.util.Values.getInt(propertyDef.substring(1));
			try
			{
				ExtraValue ev = ExtraValue.getById(dc, extraId);
				e = Dynamic.extraValue(ev);
			}
			catch (ItemNotFoundException ex)
			{}
		}
		else if (propertyDef.startsWith("="))
		{
			// propertyDef is JEP expression
			JepFunction[] jepFunctions = null;
			if (query instanceof AbstractSqlQuery)
			{
				AbstractSqlQuery dquery = (AbstractSqlQuery)query;
				jepFunctions = dquery.getJepFunctions(dc, false);	
			}
			e = Jep.formulaToExpression(propertyDef.substring(1), jepFunctions);
		}
		else
		{
			// propertyDef is a virtual column
			VirtualColumn column = null;
			if ("POSITION".equals(propertyDef))
			{
				column = VirtualColumn.POSITION;
			}
			else if ("COLUMN".equals(propertyDef))
			{
				column = VirtualColumn.COLUMN;
			}
			if (column != null) e = Dynamic.column(column);
		}
		return e;
	}
	
	/**
		@deprecated In 3.9.2, use {@link #getDynamicSelect(DbControl, SqlQuery, String)} instead
	*/
	@Deprecated
	public static Select getDynamicSelect(DbControl dc, String propertyDef)
	{
		return getDynamicSelect(dc, null, propertyDef);
	}

	/**
		Same as {@link #getDynamicExpression(DbControl, String)} but generates
		a select object instead.
	 	@param dc DbControl used to access the database 
	 	@param propertyDef The string to use for creating the expression.
	 	@return An Expression object.
	 	@since 3.9.2
	*/
	public static Select getDynamicSelect(DbControl dc, SqlQuery query, String propertyDef)
	{
		Select s = null;
		if (propertyDef.startsWith("%"))
		{
			// propertyDef is $channel_number
			int channel = Integer.parseInt(propertyDef.substring(1));
			s = Dynamic.select(VirtualColumn.channelRaw(channel));
		}
		else if (propertyDef.startsWith("$"))
		{
			// propertyDef is $rawDataPropertyName
			s = Dynamic.selectRawData(propertyDef.substring(1));
		}
		else if (propertyDef.startsWith("@@"))
		{
			// propertyDef is @@reporterPropertyName (from master table)
			s = Dynamic.selectReporter(propertyDef.substring(2), false);
		}
		else if (propertyDef.startsWith("@"))
		{
			// propertyDef is @reporterPropertyName
			s = Dynamic.selectReporter(propertyDef.substring(1));
		}
		else if (propertyDef.startsWith("#"))
		{
			// propertyDef id of extra value: #id
			int extraId = net.sf.basedb.util.Values.getInt(propertyDef.substring(1));
			try
			{
				ExtraValue ev = ExtraValue.getById(dc, extraId);
				s = Dynamic.select(ev, "xtra" + extraId);
			}
			catch (ItemNotFoundException ex)
			{}
		}
		else if (propertyDef.startsWith("="))
		{
			// propertyDef is JEP expression
			JepFunction[] jepFunctions = null;
			if (query instanceof AbstractSqlQuery)
			{
				AbstractSqlQuery dquery = (AbstractSqlQuery)query;
				jepFunctions = dquery.getJepFunctions(dc, false);	
			}
			Expression e = Jep.formulaToExpression(propertyDef.substring(1), jepFunctions);
			s = Selects.expression(e, "jep" + dc.uniqueRandom());
		}
		else
		{
			// propertyDef is a virtual column
			VirtualColumn column = null;
			if ("POSITION".equals(propertyDef))
			{
				column = VirtualColumn.POSITION;
			}
			else if ("COLUMN".equals(propertyDef))
			{
				column = VirtualColumn.COLUMN;
			}
			if (column != null) s = Dynamic.select(column);
		}
		return s;
	}
	
	/**
		Enumeration used to specify the sort direction.
	*/
	public enum SortDirection
	{
		/**
			Items are sorted in ascending order.
		*/
		ASC(0, "ascending")
		{
			@Override
			public Order sortBy(Expression e)
			{
				return Orders.asc(e);
			}
			@Override
			public SortDirection getReverseDirection()
			{
				return DESC;
			}
		},
		
		/**
			Items are sorted in descending order.
		*/
		DESC(1, "descending")		
		{
			@Override
			public Order sortBy(Expression e)
			{
				return Orders.desc(e);
			}
			@Override
			public SortDirection getReverseDirection()
			{
				return ASC;
			}
		};
	
		/**
			Get the <code>SortDirection</code> object when you know the integer code.
			
			@param value The integer value.
			@return The location for the integer value.
		*/
		public static SortDirection fromValue(int value)
		{
			SortDirection sd = valueMapping.get(value);
			assert sd != null : "sortDirection == null for value "+value;
			return sd;
		}
		
		/**
			Maps an integer to a sort direction.
		*/
		private static final Map<Integer, SortDirection> valueMapping = new HashMap<Integer, SortDirection>();
		
		static
		{
			for (SortDirection sortDirection : SortDirection.values())
			{
				SortDirection sd = valueMapping.put(sortDirection.getValue(), sortDirection);
				assert sd == null : "Another direction with the value "+sortDirection.getValue()+" already exists";
			}
		}
		
		/**
			The integer value of this direction. 
		*/
		private final int value;
		/**
			The string value of this direction.
		*/
		private final String displayValue;
	
		/**
			Creates a new direction. Used internal only.
			
			@param value The integer value of the direction
			@param displayValue Returned by the toString method
		*/
		private SortDirection(int value, String displayValue)
		{
			this.value = value;
			this.displayValue = displayValue;
		}
		
		/**
			Create an order expression that can be used with
			queries: {@link Query#order(Order)}.
			@param e The expression to sort by
			@return An order object
			@since 2.13
		*/
		public abstract Order sortBy(Expression e);
		
		/**
			Get the reverse sort direction.
			@since 3.3
		*/
		public abstract SortDirection getReverseDirection();
		
		@Override
		public String toString()
		{
			return displayValue;
		}
		
		/**
			Get the integer value that is used when storing an location to the database.
			
			@return The integer value for this location.
		*/
		public int getValue()
		{
			return value;
		}
	}

	/**
		Holds a pair of filter. One primary and one temporary	filter.
	*/
	static class PropertyFilterPair
	{
		PropertyFilter primary;
		PropertyFilter temporary;
		
		PropertyFilterPair()
		{}
		
		PropertyFilter getEffectiveFilter()
		{
			return temporary != null ? temporary : primary;
		}
		
	}
	
}
