/**
	$Id: SchemaGenerator.java 7455 2018-03-06 09:12:30Z nicklas $

	Copyright (C) 2010 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.core.hibernate;

import java.sql.Connection;
import java.sql.DatabaseMetaData;
import java.sql.SQLException;
import java.sql.Statement;
import java.util.ArrayList;
import java.util.Collections;
import java.util.EnumSet;
import java.util.HashMap;
import java.util.Iterator;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.DatabaseException;
import net.sf.basedb.core.ProgressReporter;
import net.sf.basedb.core.StringUtil;
import net.sf.basedb.core.dbengine.DbEngine;
import net.sf.basedb.core.dbengine.TableInfo;
import net.sf.basedb.util.EqualsHelper;

import org.hibernate.boot.Metadata;
import org.hibernate.boot.model.naming.Identifier;
import org.hibernate.boot.model.relational.QualifiedTableName;
import org.hibernate.boot.spi.MetadataImplementor;
import org.hibernate.dialect.Dialect;
import org.hibernate.engine.jdbc.env.internal.QualifiedObjectNameFormatterStandardImpl;
import org.hibernate.engine.jdbc.env.spi.JdbcEnvironment;
import org.hibernate.engine.jdbc.env.spi.QualifiedObjectNameFormatter;
import org.hibernate.jdbc.Work;
import org.hibernate.mapping.Table;
import org.hibernate.service.ServiceRegistry;
import org.hibernate.service.spi.ServiceRegistryImplementor;
import org.hibernate.tool.schema.SourceType;
import org.hibernate.tool.schema.TargetType;
import org.hibernate.tool.schema.extract.spi.DatabaseInformation;
import org.hibernate.tool.schema.extract.spi.TableInformation;
import org.hibernate.tool.schema.internal.SchemaCreatorImpl;
import org.hibernate.tool.schema.internal.exec.JdbcContext;
import org.hibernate.tool.schema.internal.ExceptionHandlerHaltImpl;
import org.hibernate.tool.schema.internal.Helper;
import org.hibernate.tool.schema.internal.HibernateSchemaManagementTool;
import org.hibernate.tool.schema.spi.ExceptionHandler;
import org.hibernate.tool.schema.spi.ExecutionOptions;
import org.hibernate.tool.schema.spi.SchemaCreator;
import org.hibernate.tool.schema.spi.ScriptSourceInput;
import org.hibernate.tool.schema.spi.ScriptTargetOutput;
import org.hibernate.tool.schema.spi.SourceDescriptor;
import org.hibernate.tool.schema.spi.TargetDescriptor;

/**
	Replaces the Hibernate SchemaExport and SchemaUpdate tools. We need
	this to be able to intercept and modify some of the SQL statments that
	Hibernate generates since they don't always work as expected.
	
	* Since Hibernate 5 we only trust table and column definitions. Foreign keys
	and indexes are created with {@link DbIndexWork}.
	
	* Hibernate generate "not null" constraints for
	columns that are added to existing tables and we don't want that.

	@author Nicklas
	@since 2.16
	@base.modified $Date: 2018-03-06 10:12:30 +0100 (ti, 06 mar 2018) $
*/
public class SchemaGenerator
	implements Work
{
	
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(SchemaGenerator.class);
	
	private final Metadata metadata;
	private final Dialect dialect;
	private final DbEngine dbEngine;
	private final Mode mode;
	private final ProgressReporter progress;
	private List<String> ignoredSql;
	
	/**
		Create a new schema generator. 
		
		@param metadata Information about current configuration
		@param dialect The dialect of the database we are connecting to
		@param dbEngine The DbEngine for the database we are connecting to
		@param mode The installation mode
		@param progress An optional progress reporter
	*/
	public SchemaGenerator(Metadata metadata, Dialect dialect, DbEngine dbEngine, Mode mode, ProgressReporter progress)
	{
		this.metadata = metadata;
		this.dialect = dialect;
		this.dbEngine = dbEngine;
		this.mode = mode;
		this.progress = progress;
	}

	/*
		From the Work interface
		-----------------------
	*/
	@SuppressWarnings("unchecked")
	@Override
	public void execute(Connection connection) 
		throws SQLException
	{
		log.info("Starting schema generator");
		log.info("Dialect: " + dialect);
		log.info("DbEngine: " + dbEngine);
		log.info("Mode: " + mode);
		
		this.ignoredSql = new ArrayList<String>();
		
		Statement stmt = connection.createStatement();
		String currentSql = null;
		boolean autoCommit = connection.getAutoCommit();

		try
		{
			final CollectSqlTarget schemaUpdates = new CollectSqlTarget(dbEngine, dialect, mode, connection);
			final CollectSqlTarget indexUpdates = new CollectSqlTarget();
			
			ServiceRegistry serviceRegistry = ((MetadataImplementor)metadata).getMetadataBuildingOptions().getServiceRegistry();
			HibernateSchemaManagementTool tool = new HibernateSchemaManagementTool();
			tool.injectServices((ServiceRegistryImplementor)serviceRegistry);
			
			if (mode == Mode.UPDATE)
			{
				log.info("Generating schema update script");
				
				DatabaseMetaData jdbcMetaData = connection.getMetaData();
				JdbcContext jdbcContext = tool.resolveJdbcContext(schemaUpdates.getConfigurationValues());
				JdbcEnvironment jdbcEnvironment = metadata.getDatabase().getJdbcEnvironment();
				
				QualifiedObjectNameFormatter tableNameFormatter = new QualifiedObjectNameFormatterStandardImpl(jdbcEnvironment.getNameQualifierSupport());
				
				DatabaseInformation hibernateDbInfo = Helper.buildDatabaseInformation(serviceRegistry, 
						tool.getDdlTransactionIsolator(jdbcContext), 
						metadata.getDatabase().getDefaultNamespace().getName());
				
				Iterator<Table> tables = metadata.collectTableMappings().iterator();
				while (tables.hasNext())
				{
					Table table = tables.next();
					TableInfo jdbcTableInfo = new TableInfo(table, jdbcMetaData);
					
					log.debug("Table: " + table + (jdbcTableInfo.existsInDb() ? " EXISTS" : "NOT FOUND"));
					
					if (jdbcTableInfo.existsInDb())
					{
						// This table exists in the database, we need to force Hibernate to check and ALTER it if needed
						QualifiedTableName qft = table.getQualifiedTableName();
						
						// First try: let Hibernate use the default settings to find the table
						TableInformation hibernateTableInfo = hibernateDbInfo.getTableInformation(qft);
						
						if (log.isDebugEnabled()) log.debug(qft + ": " + hibernateTableInfo);
						
						if (hibernateTableInfo == null)
						{
							// If that fails, we need to try again with unquoted table name
							qft = new QualifiedTableName(qft.getCatalogName(), qft.getSchemaName(), Identifier.toIdentifier(table.getName(), false));
							hibernateTableInfo = hibernateDbInfo.getTableInformation(qft);
							if (log.isDebugEnabled()) log.debug(qft + ": " + hibernateTableInfo);
							
							// If this is null here we are in trouble!!!
							if (hibernateTableInfo == null)
							{
								throw new DatabaseException("Table '" + table.getName() + "' exists, but Hibernate can't find any information about it. "+
										"Please report as a bug to the BASE developers");
							}
						}
						
						// To get the quoting, upper- and lower-case correct we need to re-create a new instance
						// and do a search-and-replace in the SQL generated by Hibernate
						qft = new QualifiedTableName(hibernateTableInfo.getName().getCatalogName(), hibernateTableInfo.getName().getSchemaName(), table.getNameIdentifier());
						String correctTableName = tableNameFormatter.format(qft, dialect);
						String tableNameGeneratedByHibernate = tableNameFormatter.format(hibernateTableInfo.getName(), dialect);
						if (log.isDebugEnabled())
						{
							log.debug("Search-and-replace: " + tableNameGeneratedByHibernate + "-->" + correctTableName);
						}
						Iterator<String> it = (Iterator<String>)table.sqlAlterStrings(dialect, metadata, hibernateTableInfo, null, null);
						while (it.hasNext())
						{
							String sql = it.next();
							String modifiedSql = sql.replace(tableNameGeneratedByHibernate, correctTableName);
							if (!sql.equals(modifiedSql))
							{
								log.debug("Hibernate SQL: " + sql);
								log.debug("Modified  SQL: " + modifiedSql);
							}
							schemaUpdates.accept(modifiedSql);
						}
					}
					else
					{
						String[] sql = dialect.getTableExporter().getSqlCreateStrings(table, metadata);
						for (String s : sql)
						{
							schemaUpdates.accept(s);
						}
					}	
				}
			}
			else if (mode == Mode.INSTALL)
			{
				log.info("Generating schema creation script");
				SchemaCreator schemaCreator = new SchemaCreatorImpl(tool);
				schemaCreator.doCreation(metadata, schemaUpdates, schemaUpdates, schemaUpdates);
			}
			else
			{
				throw new UnsupportedOperationException(mode.name());
			}
			
			// Collect SQL statements that are required for creating foreign keys and indexes
			DbIndexWork dbIndex = new DbIndexWork(metadata, dialect, dbEngine, false, true, false, false);
			dbIndex.setTarget(indexUpdates);
			dbIndex.execute(connection);
			
			List<SqlStatement> allSql = new ArrayList<SqlStatement>();
			allSql.addAll(schemaUpdates.getSql());// May contain null statements
			allSql.addAll(indexUpdates.getSql());

			int numStatements = 
				schemaUpdates.getNumStatements()+
				indexUpdates.getNumStatements(); // Only counts non-null statements
			int numDone = 0;
			int interval = Math.max(1, numStatements / 30);
			
			for (SqlStatement sql : allSql)
			{
				if (sql.modified != null)
				{
					if (sql.isModified())
					{
						log.debug("Original SQL: " + sql.original);
					}
					log.debug("Executing: " + sql.modified);
					
					currentSql = sql.modified;
					stmt.executeUpdate(sql.modified);
					currentSql = null;

					if (progress != null)
					{
						if (numDone % interval == 0)
						{
							progress.display((100 * numDone) / numStatements, "--" + StringUtil.trimString(sql.modified, 55));
						}
						else
						{
							progress.append(".");
						}
					}
					numDone++;
				}
				else
				{
					ignoredSql.add(sql.original);
					log.debug("Ignoring: " + sql.original);
				}
			}
		}
		catch (SQLException | RuntimeException ex)
		{
			if (currentSql != null)
			{
				log.error("Schema update failed while executing: " + currentSql, ex);
			}
			else
			{
				log.error("Schema update failed", ex);
			}
			throw ex;
		}
		finally
		{
			if (autoCommit != connection.getAutoCommit()) connection.setAutoCommit(autoCommit);
			if (stmt != null) 
			{
				try
				{
					stmt.close();
				}
				catch (SQLException ex)
				{}
			}
		}
		log.info("Schema update complete" );
		
	}

	
	/**
		Get a list with the SQL statements that was ignored in the last run.
		@return A list or null if the generator hasn't been executed
		@since 3.1
	*/
	public List<String> getIgnoredSql()
	{
		return ignoredSql;
	}
	
	/**
		Helper class used for collecting SQL statements generated by Hibernate
		schema creation/update tools.
		@since 3.8
	*/
	static class CollectSqlTarget
		implements ScriptTargetOutput, SourceDescriptor, TargetDescriptor, ExecutionOptions
	{
		private final List<SqlStatement> allSql;
		private final Dialect dialect;
		private final DbEngine dbEngine;
		private final Map<Object, Object> config;
		private final Mode mode;
		private int numStatements;
		
		CollectSqlTarget()
		{
			this(null, null, null, null);
		}
		
		CollectSqlTarget(DbEngine dbEngine, Dialect dialect, Mode mode, Connection connection)
		{
			this.dbEngine = dbEngine;
			this.dialect = dialect;
			this.mode = mode;
			// The supplied connection is passed on to Hibernate tools via the "javax...." configuration setting
			Map<Object, Object> tmp = new HashMap<>();
			if (connection != null) tmp.put("javax.persistence.schema-generation-connection", connection);
			this.config = Collections.unmodifiableMap(tmp);
			this.allSql = new ArrayList<SqlStatement>();
			this.numStatements = 0;
		}
		
		public List<SqlStatement> getSql()
		{
			return allSql;
		}
		
		public int getNumStatements()
		{
			return numStatements;
		}
		
		@Override
		public void release() 
		{}
		
		@Override
		public void prepare() 
		{}
		
		@Override
		public void accept(String sql) 
		{
			String modifiedSql = dbEngine == null ? sql : dbEngine.inspectSchemaGenerationSQL(sql, dialect, mode);
			allSql.add(new SqlStatement(sql, modifiedSql));
			if (modifiedSql != null) numStatements++;
		}

		@Override
		public ScriptSourceInput getScriptSourceInput() 
		{
			return null;
		}

		@Override
		public SourceType getSourceType()
		{
			return SourceType.METADATA;
		}
		
		@Override
		public EnumSet<TargetType> getTargetTypes() 
		{
			return EnumSet.of(TargetType.SCRIPT);
		}

		@Override
		public ScriptTargetOutput getScriptTargetOutput() 
		{
			return this;
		}

		@Override
		public Map<?, ?> getConfigurationValues() 
		{
			return config;
		}

		@Override
		public boolean shouldManageNamespaces()
		{
			return false;
		}

		@Override
		public ExceptionHandler getExceptionHandler() 
		{
			return ExceptionHandlerHaltImpl.INSTANCE;
		}
		
		
	}
	
	public static class SqlStatement
	{
		final String original;
		final String modified;
		
		public SqlStatement(String sql)
		{
			this(sql, sql);
		}
		
		public SqlStatement(String original, String modified)
		{
			this.original = original;
			this.modified = modified;
		}
		
		boolean isModified()
		{
			return !EqualsHelper.equals(original, modified);
		}
	}
	
	/**
		The installation mode.
		@since 3.1
	*/
	public static enum Mode
	{
		/**
			Creating a fresh installation.
		*/
		INSTALL(true, false),
		/**
			Updating an existing installation.
		*/
		UPDATE(false, true);
		
		private final boolean requireEmptyDatabase;
		private final boolean verifySchemaVersion;
		private Mode(boolean requireEmptyDatabase, boolean verifySchemaVersion)
		{
			this.requireEmptyDatabase = requireEmptyDatabase;
			this.verifySchemaVersion = verifySchemaVersion;
		}
		
		/**
			Does this installation mode require an empty database or not?
		*/
		public boolean requireEmptyDatabase()
		{
			return requireEmptyDatabase;
		}
		
		/**
			Does this installation mode require a check against the schema version
			of the intalled database or not?
		*/
		public boolean verifySchemaVersion()
		{
			return verifySchemaVersion;
		}
	}
	
	
}
