/**
	$Id $

	Copyright (C) 2017 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util;

import java.io.InputStream;
import java.io.InputStreamReader;
import java.io.StringWriter;
import java.io.UnsupportedEncodingException;
import java.net.URI;
import java.net.URLEncoder;

import org.apache.commons.validator.routines.InetAddressValidator;
import org.apache.http.Header;
import org.apache.http.HttpEntity;
import org.apache.http.HttpResponse;
import org.apache.http.HttpStatus;
import org.apache.http.StatusLine;
import org.json.simple.JSONObject;
import org.json.simple.parser.JSONParser;

import net.sf.basedb.core.AbsoluteProgressReporter;
import net.sf.basedb.core.Config;
import net.sf.basedb.core.InvalidDataException;
import net.sf.basedb.util.filter.Filter;
import net.sf.basedb.util.uri.http.HttpConnectionManager;

/**
	Utility class for geographic location.

	@author nicklas
	@since 3.12
*/
public class GeoLocation
{

	/**
		Try to convert the given ip address to a location. This will
		query the 'freegeoip' server specified in base.config. The default
		is ''. If this has been disabled or if the location is unknown,
		null is returned.
	*/
	@SuppressWarnings("deprecation")
	public static GeoLocation getLocationFromIp(String ip)
	{
		if (!InetAddressValidator.getInstance().isValid(ip))
		{
			throw new InvalidDataException("Invalid ip address: " + ip);
		}
		
		String url = Config.getString("geolocation.freegeoip");
		if (url == null) return null;
		
		try
		{
			url = url.replace("{IP}", URLEncoder.encode(ip, "UTF-8"));
		}
		catch (UnsupportedEncodingException ex)
		{
			url = url.replace("{IP}", URLEncoder.encode(ip));
		}
		
		JSONObject jsonLoc = null;
		InputStream in = null;
		try
		{
			ResponseFilter filter = new ResponseFilter();
			HttpConnectionManager http = new HttpConnectionManager(new URI(url), null);
			http.setResponseFilter(filter);
			http.setTimeout(3); // Do not allow more than 3 seconds!
			in = http.getInputStream();
			StringWriter sout = new StringWriter();
			FileUtil.copy(new InputStreamReader(in, "UTF-8"), sout, filter);
			jsonLoc = (JSONObject)new JSONParser().parse(sout.toString());
			in.close();
		}
		catch (Exception ex)
		{
			FileUtil.close(in);
			throw new RuntimeException(url, ex);
		}
		
		return fromJson(jsonLoc);
	}
	
	/**
		Get an URL that opens a map marking the given position. A
		link template must have been configured in base.config (geolocation.maptemplate).
		Use {LAT} and {LONG} as placeholders for the latitude
		and longitude.

		@param latitude The latitude of the location
		@param longitude The longitude of the location
		@return A link or null if either of the parameters is null
	*/
	public static String getURLToMap(Float latitude, Float longitude)
	{
		if (latitude == null || longitude == null) return null;
		String template = Config.getString("geolocation.maptemplate");
		if (template == null) return null;
		String url = template.replace("{LAT}", latitude.toString()).replace("{LONG}", longitude.toString());
		return url;
	}
	
	/**
		Convert JSON data from freegeoip to a GeoLocation instance.
		The JSON must contain at least one of "city", "region_name" or 
		"country_name".
		@return A location object or null if there is no data
	*/
	public static GeoLocation fromJson(JSONObject json)
	{
		String city = Values.getStringOrNull((String)json.get("city"));
		String region = Values.getStringOrNull((String)json.get("region_name"));
		String country = Values.getStringOrNull((String)json.get("country_name"));
		if (city == null && region == null && country == null) return null;
		
		GeoLocation loc = new GeoLocation();
		loc.setCity(city);
		loc.setRegion(region);
		loc.setCountry(country);
		Number latitude = (Number)json.get("latitude");
		Number longitude = (Number)json.get("longitude");
		if (latitude != null) loc.setLatitude(latitude.floatValue());
		if (longitude != null) loc.setLongitude(longitude.floatValue());
		return loc;
	}
		
	private String city;
	private String region;
	private String country;
	
	private Float latitude;
	private Float longitude;
	
	public GeoLocation()
	{}
	
	public String getCity() 
	{
		return city;
	}

	public void setCity(String city)
	{
		this.city = city;
	}

	public String getRegion() 
	{
		return region;
	}

	public void setRegion(String region) 
	{
		this.region = region;
	}

	public String getCountry() 
	{
		return country;
	}

	public void setCountry(String country) 
	{
		this.country = country;
	}

	public Float getLatitude() 
	{
		return latitude;
	}

	public void setLatitude(Float latitude) 
	{
		this.latitude = latitude;
	}

	public Float getLongitude() 
	{
		return longitude;
	}

	public void setLongitude(Float longitude) 
	{
		this.longitude = longitude;
	}

	@Override
	public String toString()
	{
		StringBuilder sb = new StringBuilder();
		append(sb, city);
		append(sb, region);
		append(sb, country);
		return sb.length() == 0 ? null : sb.toString();
	}
	
	private void append(StringBuilder sb, String s)
	{
		if (s == null) return;
		if (sb.length() > 0) sb.append(", ");
		sb.append(s);
	}
	
	/**
		We only accept "200" status response, "application/json" content type
		and not more than 1K bytes (should be a lot less under normal operations).
	*/
	static class ResponseFilter
		implements Filter<HttpResponse>, AbsoluteProgressReporter
	{

		@Override
		public boolean evaluate(HttpResponse response) 
		{
			StatusLine status = response.getStatusLine();
			if (status != null && status.getStatusCode() != HttpStatus.SC_OK)
			{
				throw new RuntimeException(status.toString());
			}
			
			HttpEntity entity = response.getEntity();
			Header contentType = entity.getContentType();
			if (contentType != null && !"application/json".equals(contentType.getValue()))
			{
				throw new RuntimeException("Unexpected content type: "+contentType.getValue());
			}
			
			long size = entity.getContentLength();
			if (size > 1000)
			{
				throw new RuntimeException("Too much data: " + size);
			}
			
			return true;
		}

		@Override
		public void display(int percent, String message) 
		{}

		@Override
		public void append(String message) 
		{}

		@Override
		public void displayAbsolute(long completed, String message) 
		{
			if (completed > 1000)
			{
				throw new RuntimeException("Too much data: " + completed);
			}
		}
		
	}
}
