/**
	$Id: ExtensionsFile.java 7273 2017-01-20 09:35:05Z nicklas $

	Copyright (C) 2011 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.extensions.manager;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.InputStream;
import java.net.URI;
import java.util.ArrayList;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;
import java.util.concurrent.TimeUnit;
import java.util.concurrent.locks.ReentrantReadWriteLock;
import java.util.zip.ZipEntry;
import java.util.zip.ZipFile;

import net.sf.basedb.core.plugin.About;
import net.sf.basedb.util.FileUtil;
import net.sf.basedb.util.JarClassLoader;
import net.sf.basedb.util.extensions.Extension;
import net.sf.basedb.util.extensions.ExtensionPoint;
import net.sf.basedb.util.extensions.xml.XmlLoader;
import net.sf.basedb.util.uri.CloseResourceInputStream;

/**
	Represents a file with extensions in it. The file is either an XML file
	or a JAR file. A JAR file is always a real file on the file system and
	can be accessed with {@link #getFile()}. An XML file can be both a real file
	or an abstract stream of bytes and can be accessed by {@link #getXmlStream()}.
	The latter method can also be used in the case of a JAR file and will then
	open the <code>META-INF/extensions.xml</code> file.


	@author Nicklas
	@since 3.0
	@base.modified $Date: 2017-01-20 10:35:05 +0100 (fr, 20 jan 2017) $
 */
public class ExtensionsFile
	implements Comparable<ExtensionsFile>
{
	private static final org.slf4j.Logger log = 
		org.slf4j.LoggerFactory.getLogger(ExtensionsFile.class);

	private final ExtensionsManager manager;
	private final File file;
	private final URI uri;
	private final boolean isJar;
	private final String name;
	private final ReentrantReadWriteLock rwLock;
	private final Map<ObjectKey<?>, Object> allObjects;
	private final Map<ObjectKey<?>, Object> objectMetadata;
	
	private boolean isNew;
	private boolean wasModified;
	private long lastModified;
	private long lastLength;
	private JarClassLoader jarLoader;
	private About about;
	private boolean hasError;

	private volatile boolean isValid;
	private Throwable validationError;

	/**
		Create a new extensions file backed by a real file on the
		file system. If the file name ends with '.jar' it is assumed
		to be a JAR file, otherwise it must be an XML file containing
		extension definitions.
		@param file The file
	*/
	ExtensionsFile(ExtensionsManager manager, File file)
	{
		this(manager, file, file.getName().endsWith(".jar"));
	}
	
	/**
		Create a new extensions file backed by a real file
		on the file system. The file may be a JAR file or an
		XML file.
		
		@param file The file
		@param isJar TRUE if it is a JAR file, FALSE if it is an XML file
	 */
	ExtensionsFile(ExtensionsManager manager, File file, boolean isJar)
	{
		this(manager, file, file.toURI(), file.getName(), isJar);
	}

	/**
		Create a new extension file backed by an URI. The URI must point to
		an XML file containing extension definitions.
		@param uri The URI
	*/
	ExtensionsFile(ExtensionsManager manager, URI uri)
	{
		this(manager, null, uri, uri.toString().replaceFirst(".*/", ""), false);
	}
	
	private ExtensionsFile(ExtensionsManager manager, File file, URI uri, String name, boolean isJar)
	{
		this.manager = manager;
		this.file = file;
		this.uri = uri;
		this.name = name;
		this.isJar = isJar;
		this.isNew = true;
		this.wasModified = true;
		this.rwLock = new ReentrantReadWriteLock();
		this.allObjects = new LinkedHashMap<ObjectKey<?>, Object>();
		this.objectMetadata = new HashMap<ObjectKey<?>, Object>();
	}
	
	/*
		From the Comparable interface
		-----------------------------
	*/
	@Override
	public int compareTo(ExtensionsFile other)
	{
		// Compare by name and then by URI
		int n = this.name.compareTo(other.name);
		return n == 0 ? this.uri.compareTo(other.uri) : n;
	}
	// ------------------------------
	/*
		From the Object class
		---------------------
	*/
	@Override
	public int hashCode()
	{
		return uri.hashCode();
	}
	@Override
	public boolean equals(Object o)
	{
		if (this == o) return true;
		if (o == null || this.getClass() != o.getClass()) return false;
		ExtensionsFile other = (ExtensionsFile)o;
		return this.uri.equals(other.uri);
	}
	@Override
	public String toString()
	{
		return "ExtensionsFile@" + System.identityHashCode(this) + "[" + uri + "]";
	}
	// ------------------------------
	
	/**
		Get the name of the file. The name may or may not
		correspond to a real file.
	*/
	public String getName()
	{
		return name;
	}
	
	/**
		Get the underlying file on the file system. Return null
		if the extension is defined by an URI.
	*/
	public File getFile()
	{
		return file;
	}
	
	/**
		Get an URI that points to the extensions file.
	*/
	public URI getURI()
	{
		return uri;
	}
	
	/**
		If the file name ends with '.jar' the file is considered a JAR file.
	 	Otherwise it is considered an XML file. The file name check is case
	 	sensitive.
		@return TRUE if the file is a JAR file, false if it is an XML file
	*/
	public boolean isJar()
	{
		return isJar;
	}
	
	/**
		Check if the file exists. For files that are only represented by
		an URI this method always return true. For other files the call 
		is forwarded to {@link java.io.File#exists()} and {@link java.io.File#isFile()}.
		@return See {@link java.io.File#exists()}
	*/
	public boolean exists()
	{
		return file == null || (file.exists() && file.isFile());
	}

	/**
		If the file is a new file not previously processed by the
		extension system. Thew 'new' status is changed when 
		{@link WriteableExtensionsFile#markAsProcessed()} is called which 
		usually happens when all extensions has been registered.
		@return TRUE if the file is a new file, FALSE otherwise
	*/
	public boolean isNew()
	{
		return isNew;
	}

	/**
		Is this file installed into the system or not?
	*/
	public boolean isInstalled()
	{
		return file == null || manager.getSettings().isInstalledFile(file);
	}
	
	/**
		Is this an extension file that should be ignored?
		@since 3.10
	*/
	public boolean isIgnored()
	{
		return file != null && manager.getSettings().isIgnoredFile(file);
	}

	/**
		Check if the file was modified when the last call to
		{@link #checkModified()} was made. It is recommended
		that processor implementation use this method instead
		to avoid inconsistent behaviour if the file happens to
		be modified while a processor is running.
		
		@return TRUE if the file was modified, FALSE otherwise
	*/
	public boolean wasModified()
	{
		return wasModified;
	}
	
	/**
		Check if the underlying file has been modified since it was
		last processed. Files that are new are always considered to be
		modified. Files that are only represented as an URI are considered
		to be unmodifiable. For other files, we check the 
		{@link java.io.File#lastModified()} and {@link java.io.File#length()} of the 
		underlying file and compare that to the last known timestamp and size. 
		Files that are JAR files we also check with the class loader if any of
		the libraries it uses have changed.
		<p>
		The result of this call is remembered until this method is called 
		again or until the file is marked as processed.
		
		@see WriteableExtensionsFile#markAsProcessed()
		@see #wasModified()
	*/
	public boolean checkModified()
	{
		log.debug("Checking if file is modified: " + uri);
		wasModified = false; // Assume not modified as a starting point
		if (isNew)
		{
			// New files are always modified
			log.debug("File was new: " + uri);
			wasModified = true;
		}
		else if (file != null && (lastModified != file.lastModified() || lastLength != file.length()))
		{
			// The XML or JAR file has changed
			log.debug("File has been modified: " + uri);
			wasModified = true;
		}
		else if (isJar && jarLoader != null && jarLoader.hasChanged(true))
		{
			// JAR files have changed if the class loader detect changes
			log.debug("Class path has been modified: " + uri);
			wasModified = true;
		}
		else
		{
			log.debug("File has not been modified: " + uri);
		}
		return wasModified;
	}
	
	/**
		Is the file a valid extensions file according to the last
		performed validation. 
		
		This means that the file must be either:
		<ul>
		<li>An extensions definition XML file 
		<li>A JAR file with an extensions definition XML file at
			<code>META-INF/extensions.xml</code>.
		</ul>
		
		The result of the validation is remembered as long as the 
		file is not modified. If {@link #checkModified()} returns
		true, the return value of this method may be out of sync.
		Re-validation of modified files are usually performed
		by the extensions manager by calling
		{@link ExtensionsManager#scanForChanges()}
		
		<p>
		NOTE! The validation will only verify that the XML file 
		follows the rules defined by the schema definition. It will
		not check that actual classes, etc. exists and can be created.
		This usually happens later in the registration process, and
		can be checked by {@link #hasError()}.
		
		@return TRUE if the file is valid, FALSE otherwise
		@see #getValidationError()
		@see #checkModified()
		@see #hasError()
	*/
	public boolean isValid()
	{
		return isValid;
	}

	/**
		Get more information about the error that caused the validation
		to fail.
		@return An exception or null if the validation succeeded
		@see #isValid()
	*/
	public Throwable getValidationError()
	{
		return validationError;
	}
	
	/**
		Get information about the extensions in this file.
		@return An About object or null if the file is not a valid
			extensions file
	*/
	public About getAbout()
	{
		return about;
	}

	/**
		Get an input stream for reading the XML file containing the extension
		definitions.
		@return An input stream
	*/
	public InputStream getXmlStream()
		throws IOException
	{
		InputStream in = null;
		if (isJar())
		{
			ZipFile zipFile = null;
			try
			{
				zipFile = new ZipFile(getFile());
				ZipEntry zipEntry = zipFile.getEntry("META-INF/extensions.xml");
				if (zipEntry == null)
				{
					throw new FileNotFoundException(name+"!META-INF/extensions.xml");
				}
				in = new CloseResourceInputStream(zipFile.getInputStream(zipEntry), zipFile);
			}
			finally
			{
				if (in == null) FileUtil.close(zipFile);
			}
		}
		else
		{
			in = uri.toURL().openStream();
		}
		return in;
	}

	/**
		Get an input stream for reading the resource specified by the
		given path
		@param path The path of the resource
		@return An input stream (or null if this extensions file is not a JAR file or if the given path is not found)
	*/
	public InputStream getStream(String path)
		throws IOException
	{
		InputStream in = null;
		if (isJar())
		{
			ZipFile zipFile = null;
			try
			{
				zipFile = new ZipFile(getFile());
				ZipEntry zipEntry = zipFile.getEntry(path);
				if (zipEntry != null)
				{
					in = new CloseResourceInputStream(zipFile.getInputStream(zipEntry), zipFile);
				}
			}
			finally
			{
				if (in == null) FileUtil.close(zipFile);
			}
		}
		return in;
	}

	
	/**
		Validate the XML file with the extension definitions.
	*/
	synchronized boolean validate()
	{
		log.info("Validating extensions in file: " + this);
		isValid = false;
		hasError = false;
		validationError = null;
		
		InputStream in = null;
		try
		{
			in = getXmlStream();
			about = new XmlLoader().validateXmlFile(in, getName());
			log.info("Successfully validated extensions in file: " + this);
			isValid = true;
		}
		catch (Throwable t)
		{
			log.error("Error validating extensions in file: " + this, t);
			validationError = t;
		}
		finally
		{
			FileUtil.close(in);
		}
		return isValid;
	}

	
	/**
		If there was an error when registering the extensions in this file.
		This property is only set if the file has been determined to be a 
		valid extensions file by the {@link #isValid()} method. Extensions that
		has an error are automatically disabled.
		
		@return TRUE if there was some error, FALSE if everything is ok
	*/
	public boolean hasError()
	{
		return hasError || !isValid;
	}
	
	/**
		Get a list of all objects defined in this extensions
		file. This list can contain any type of objects that
		have been registered by a processer, but typically
		this list contains {@link ExtensionPoint}:s and
		{@link Extension}:s. Use {@link #getObjectsOfClass(Class)}
		to get a list with only one type of objects.
		
		@return A list with the objects
	*/
	public List<Object> getAllObjects()
	{
		List<Object> copy = new ArrayList<Object>();
		try
		{
			if (readLock())
			{
				copy.addAll(allObjects.values());
			}
		}
		finally
		{
			rwLock.readLock().unlock();
		}
		return copy;
	}
	
	/**
		Get a list of all objects defined in this extensions
		file that are of the specified class or interface.
		Note! The returned list may be empty if another
		thread is currently processing this file.
		
		@param ofClass The class/interface the objects must be
			an instance of
		@return A list with the objects, if no objects are found
			the list is empty
	*/
	public <T> List<T> getObjectsOfClass(Class<T> ofClass)
	{
		List<T> matching = new ArrayList<T>();
		try
		{
			if (readLock())
			{
				for (Object o : allObjects.values())
				{
					if (ofClass.isInstance(o)) matching.add(ofClass.cast(o));
				}
			}
		}
		finally
		{
			rwLock.readLock().unlock();
		}
		return matching;
	}
	
	/**
		Get the object that was registered for the given key.
		@param key An object key
		@return The object or null if no object was found
	*/
	@SuppressWarnings("unchecked")
	public <O> O getObjectForKey(ObjectKey<O> key)
	{
		return (O)allObjects.get(key);
	}
	
	/**
		Get metadata registered for a given given key.
		@param key An object key
		@return A metadata object, or null if no metadata was found
	*/
	@SuppressWarnings("unchecked")
	public <M> M getMetadata(ObjectKey<M> key)
	{
		return (M)objectMetadata.get(key);
	}

	/**
		Get a list of all metadata defined in this extensions
		file that are keyed with the specified class or interface.
		Note! The returned list may be empty if another
		thread is currently processing this file.
		
		@param ofClass The class/interface the metadata keys must be
			an instance of
		@return A list with the objects, if no objects are found
			the list is empty
	*/
	public <T> List<T> getMetadataKeysOfClass(Class<T> ofClass)
	{
		List<T> matching = new ArrayList<T>();
		try
		{
			if (readLock())
			{
				for (ObjectKey<?> key : objectMetadata.keySet())
				{
					if (ofClass.isInstance(key)) matching.add(ofClass.cast(key));
				}
			}
		}
		finally
		{
			rwLock.readLock().unlock();
		}
		return matching;
	}
	
	/**
		Get the class loader used to load classes for the extension. Only JAR
		files have class loaders so this method may return null. Once a class
		loader has been created it remains the same until a change has been detected
		with {@link #checkModified()} which forces the creation of a new class loader
		when this method is called.
	*/
	public ClassLoader getClassLoader()
		throws IOException
	{
		if ((jarLoader == null) && isJar())
		{
			jarLoader = (JarClassLoader)JarClassLoader.getInstance(getFile().getAbsolutePath(), true);
		}
		return jarLoader;
	}

	/**
		Get a writeable view for the current file. Note that the view
		is returned in read-only mode and to really be able to write information
		the method {@link WriteableExtensionsFile#open()} must be called.
	*/
	WriteableExtensionsFile getWriteableFile()
	{
		return new WriteableExtensionsFile(this);
	}
	
	/**
		Try to aquire a read-lock. A read-lock is needed when reading information
		that may be corrupted if another thread is currently writing information
		to this file.
		@return TRUE if the read-lock could be aquired immediately, FALSE if not
	*/
	boolean readLock()
	{
		return rwLock.readLock().tryLock();
	}
	
	/**
		Try to aquire a write-lock. A write-lock is needed by any thread that wants
		to update information about this file. The write-lock can only be held by
		a single thread at a time.
		@return TRUE if the write-lock could be aquired within 1 second, FALSE if not
	*/
	boolean writeLock()
	{
		try
		{
			return rwLock.writeLock().tryLock(1, TimeUnit.SECONDS);
		}
		catch (InterruptedException ex)
		{}
		return false;
	}
	
	/**
		An extensions file with additional methods that allows adding
		or modifying information in the underlying extensions file. This kind of 
		view is usually handed out to {@link ExtensionsFileProcessor} implementations
		as a result of calling {@link ExtensionsManager#processFiles(ExtensionsFileProcessor)}.
		<p>
		
		Note that the file starts out in read-only mode and that
		the processor has to call {@link #open()} to aquire a write-lock
		before calling any writing methods. It is recommended that
		the lock is released as soon as possible by calling {@link #close()}.
	*/
	public static class WriteableExtensionsFile
	{
		private final ExtensionsFile xtFile;
		private boolean isClosed;
		
		/**
			Creates a new writeable extensions file. 
		*/
		WriteableExtensionsFile(ExtensionsFile xtFile)
		{
			this.xtFile = xtFile;
			this.isClosed = true;
		}
		
		/**
			Get the underlying extensions file (for readin information)
		*/
		public ExtensionsFile getExtensionsFile()
		{
			return xtFile;
		}
		
		/**
		 	Close the writeable file. This will release the lock and
		 	any further write operations are disallowed.
		*/
		public void close()
		{
			if (isClosed) return;
			isClosed = true;
			xtFile.rwLock.writeLock().unlock();
		}
		
		public boolean open()
		{
			if (isClosed)
			{
				isClosed = !xtFile.writeLock();
			}
			return !isClosed;
		}
		
		/**
			Check if the file has been closed and throws an
			IllegalStateException if it has.
		*/
		private void checkClosed()
		{
			if (isClosed) 
			{
				throw new IllegalStateException("The file has been closed: " + xtFile);
			}
		}
		
		/**
			Mark the file as fully processed by the extensions manager. 
		*/
		public void markAsProcessed()
		{
			xtFile.isNew = false;
			xtFile.wasModified = false;
			if (xtFile.file != null)
			{
				xtFile.lastModified = xtFile.file.lastModified();
				xtFile.lastLength = xtFile.file.length();
			}
			if (!xtFile.isValid)
			{
				// Cleanup to avoid leaking memory
				xtFile.allObjects.clear();
				xtFile.manager.unregisterAllObjects(xtFile);
				xtFile.objectMetadata.clear();
				xtFile.jarLoader = null;
			}
		}
		
		/**
			Mark the file as uninstalled.
		*/
		public void setUninstalled()
		{
			xtFile.manager.removeFile(xtFile.file);
		}
		
		/**
			Mark the file as installed.
		*/
		public void setInstalled()
		{
			xtFile.manager.getSettings().setInstalledFile(xtFile.file);
		}
		
		/**
			Sets the error status. This method can be called even on closed
			files, since it is important to error handling.
		*/
		public void setError(boolean error)
		{
			xtFile.hasError = error;
		}
		
		/**
			Reset the classloader for this file, forcing a new
			instance in case the JAR file has been changed. 
			This method should be called before re-installing 
			an extension and the {@link ExtensionsFile#checkModified()} 
			returns true.
			@throws IllegalStateException If the file has been closed
			@since 3.10
		*/
		public void resetClassLoader()
		{
			checkClosed();
			if (xtFile.jarLoader != null)
			{
				xtFile.manager.getRegistry().unregisterEventHandlers(xtFile.jarLoader);
				xtFile.jarLoader = null;
			}
		}

		/**
			Register an object as "defined" by this extensions file.
			@param key The object key used to identify the object
			@param obj The object to register
			@throws IllegalStateException If the file has been closed
		*/
		public <O> void registerObject(ObjectKey<O> key, O obj)
		{
			checkClosed();
			xtFile.manager.registerObject(key, this.xtFile);
			xtFile.allObjects.put(key, obj);
		}
		
		/**
			Unregister an object that was "defined" by this extensions
			file.
			@param key The object key used to identify the object
			@throws IllegalStateException If the file has been closed
		*/
		public <O> void unregisterObject(ObjectKey<O> key)
		{
			checkClosed();
			xtFile.allObjects.remove(key);
			xtFile.manager.unregisterObject(key, this.xtFile);
		}
		
		/**
			Register metadata about an object. The difference
			between this method and the {@link #registerObject(ObjectKey, Object)}
			method is that the <code>registerObject</code> method
			also registers the objects with the {@link ExtensionsManager}.
			
			@param key The object key used to identify the metadata
			@param metadata The metadata to store under the key
		*/
		public <M> void registerMetadata(ObjectKey<M> key, M metadata)
		{
			checkClosed();
			xtFile.objectMetadata.put(key, metadata);
		}
		
		
		/**
			Unregister metadata about an object. 
			@param key The object key used to identify the metadata
		*/
		public <M> void unregisterMetadata(ObjectKey<M> key)
		{
			checkClosed();
			xtFile.objectMetadata.remove(key);
		}
		
	}
}
