/**
	$Id: ArraySlideLoader.java 5651 2011-06-08 13:00:33Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import net.sf.basedb.core.ArrayBatch;
import net.sf.basedb.core.ArraySlide;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.PhysicalBioAssay;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.node.NodeFactory;

/**
	Node loader implementation for array slides. The forward-loading
	direction goes from {@link ArrayBatch}:s -&gt; {@link ArraySlide}:s -&gt; 
	{@link PhysicalBioAssay}:s. The reverse-loading direction is the opposite 
	direction.

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-06-08 15:00:33 +0200 (on, 08 jun 2011) $
*/
public class ArraySlideLoader
	extends BasicItemNodeLoader<ArraySlide>
{

	public ArraySlideLoader()
	{
		super(Item.ARRAYSLIDE, ALLOW_ROOT_NODE, 
				new NameableNameGenerator<ArraySlide>("slide", "Array slide"));
	}
	
	/*
		From the NodeLoader interface
		------------------------------
	*/
	/**
		Create forward-loading array slide nodes from a given array batch node.
		The returned node is a folder-type node containing item nodes for 
		each slide that has been created by the batch.
		@return The folder node for the slides
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node batchNode)
	{
		NodeFactory<ArraySlide> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ArrayBatch batch = (ArrayBatch)batchNode.getItem(dc);
		ItemQuery<ArraySlide> query = context.initQuery(batch.getArraySlides(), "name");
		ItemResultIterator<ArraySlide> it = query.iterate(dc);
		while (it.hasNext())
		{
			ArraySlide slide = it.next();
			if (folderNode == null)
			{
				folderNode =  new Node("slides", "Slides", batchNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, slide, slide, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, batchNode, false);
		return folderNode;
	}
	
	/**
		Create a reverse-loading array slide node from the hyrbidization node.
		@return An array slide node, or null if the hybridization doesn't have a slide
	*/
	@Override
	public Node createReverseNode(DbControl dc, OverviewContext context, Node hybNode)
	{
		NodeFactory<ArraySlide> nf = getNodeFactory(dc, context);
		ArraySlide slide = null;
		boolean denied = false;
		try
		{
			slide = ((PhysicalBioAssay)hybNode.getItem(dc)).getArraySlide();
		}
		catch (PermissionDeniedException ex)
		{
			denied = true;
		}
		Node slideNode = createItemNode(nf, slide, slide, denied, 
				hybNode, ChildNodeDirection.REVERSE);
		return slideNode;
	}
	// --------------------------------
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	/**
		Loads property nodes of an array slide. 
		<ul>
		<li>Annotations: {@link AnnotationLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node slideNode)
	{
		getNodeLoader(context, Item.ANNOTATION).createPropertyNode(dc, context, slideNode);
		getNodeLoader(context, Item.ANYTOANY).createPropertyNode(dc, context, slideNode);
	}

	/**
		Loads the physical bioassay node that is related to the given array slide node.
		@see PhysicalBioAssayLoader#createForwardNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node slideNode)
	{
		getNodeLoader(context, Item.PHYSICALBIOASSAY).createForwardNode(dc, context, slideNode);
	}
	
	/**
		Loads the array batch node that is the parent of this array slide. 
		@see ArrayBatchLoader#createReverseNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadReverseChildNodes(DbControl dc, OverviewContext context, Node slideNode)
	{
		getNodeLoader(context, Item.ARRAYBATCH).createReverseNode(dc, context, slideNode);
	}
	// ---------------------------------------
}
