/**
	$Id$
	
	Copyright (C) 2012 Nicklas Nordborg
	
	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/
	
	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.
	
	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.
	
	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.loader;

import java.util.HashSet;
import java.util.List;
import java.util.Set;

import net.sf.basedb.core.BioPlate;
import net.sf.basedb.core.BioPlateEvent;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.ItemResultIterator;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;
import net.sf.basedb.util.overview.Node;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.filter.ItemTypeFilter;
import net.sf.basedb.util.overview.node.ChildNodeDirection;
import net.sf.basedb.util.overview.node.NameableNameGenerator;
import net.sf.basedb.util.overview.node.NodeFactory;


/**
	Node loader implementation for bioplate events. Events are
	forward-loading property nodes of a bioplate. The reverse 
	direction is not loading anything.
	
	@author nicklas
	@since 3.3
*/
public class BioPlateEventLoader 
	extends BasicItemNodeLoader<BioPlateEvent> 
{
	public BioPlateEventLoader()
	{
		super(Item.BIOPLATEEVENT, false, 
				new NameableNameGenerator<BioPlateEvent>("event", "Event"));
	}
	
	/*
		From the NodeLoader interface
		------------------------------
	*/

	/**
		Create a forward-loading node from a bioplate. Note that events that already
		exists in the parent chain are not loaded again to avoid infinite recursive
		loading.
		@return A folder node with the events, or null if there are no events for the bioplate
	*/
	@Override
	public Node createForwardNode(DbControl dc, OverviewContext context, Node plateNode)
	{
		BioPlate plate = (BioPlate)plateNode.getItem();
		NodeFactory<BioPlateEvent> nf = getNodeFactory(dc, context);
		Node folderNode = null;
		ItemQuery<BioPlateEvent> query = context.initQuery(plate.getEvents(), "name");
		
		// We should only load events that hasn't already been loaded in the parent chain
		List<Node> eventNodes = plateNode.findAll(new ItemTypeFilter(Item.BIOPLATEEVENT));
		if (eventNodes.size() > 0)
		{
			Set<Integer> eventIds = new HashSet<Integer>(eventNodes.size());
			for (Node n : eventNodes)
			{
				eventIds.add(n.getItem().getId());
			}
			query.restrict(Restrictions.not(Restrictions.in(Hql.property("id"), Expressions.parameter("events"))));
			query.setParameter("events", eventIds, Type.INT);
		}

		ItemResultIterator<BioPlateEvent> it = query.iterate(dc);
		while (it.hasNext())
		{
			BioPlateEvent event = it.next();
			if (folderNode == null)
			{
				folderNode = new Node("events", "Events", plateNode, ChildNodeDirection.FORWARD);
			}
			createItemNode(nf, event, event, false, folderNode, ChildNodeDirection.FORWARD);
		}
		postValidateFolder(nf, folderNode, plateNode, false);
		return folderNode;
	}	
	// -------------------------------------
	
	/*
		From the AbstractNodeLoader class
		----------------------------------
	*/
	
	/**
		Loads the bioplates that are part of the event.
		@see BioPlateLoader#createPropertyNode(DbControl, OverviewContext, Node)
	*/
	@Override
	protected void loadForwardChildNodes(DbControl dc, OverviewContext context, Node eventNode)
	{
		getNodeLoader(context, Item.BIOPLATE).createPropertyNode(dc, context, eventNode);
	}
	
	/**
		Loads property nodes of an event. 
		<ul>
		<li>Protocol: {@link ProtocolLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Kit: {@link KitLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		<li>Hardware: {@link HardwareLoader#createPropertyNode(DbControl, OverviewContext, Node)}
		</ul>
	*/
	@Override
	protected void loadPropertyChildNodes(DbControl dc, OverviewContext context, Node eventNode) 
	{
		getNodeLoader(context, Item.PROTOCOL).createPropertyNode(dc, context, eventNode);
		getNodeLoader(context, Item.KIT).createPropertyNode(dc, context, eventNode);
		getNodeLoader(context, Item.HARDWARE).createPropertyNode(dc, context, eventNode);
	}
	// ---------------------------------------

}
