/**
	$Id$

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.node;

import net.sf.basedb.util.overview.Node;


/**
	Represents the direction to use for loading child nodes in a tree.
	We need to keep track of this since we need to support rooting a tree
	at any type of item (eg. at an experiment, hybridization, biosource, etc).
	<p>
	
	For every tree there is a <i>natural</i> order of items. For example,
	BioSource -&gt; Sample -&gt; Extract -&gt; ... -&gt; Experiment.
	The natural order is the forward-loading direction and the opposite
	order is the reverse-loading direction.
	<p>
	
	For example, if we root the tree at a Hybridization we will load children 
	in the forward-loading direction down to experiment and in the 
	revers-loading direction up to biosource and array design. For each node
	that is added to a tree we need to keep track if we should continue up
	or down the natural order. Otherwise we will end up in an infinite loop.
	<p>
	
	In a tree there can also be side-nodes that in most cases are end-point or
	leaf nodes. This is for example nodes containing information about
	protocol, software, annotations, etc. We call such nodes <i>property</i>
	nodes and they are usually loaded in both the forward- and reverse-loading
	directions.
	
	@author Nicklas
	@version 2.10
	@base.modified $Date$
	@see Node#getChildNodeDirection()
*/
public enum ChildNodeDirection
{
	/**
		Represents an end-point node. No more child nodes should be loaded,
		not even property nodes.
	*/
	NONE(false, false, false),
	
	/**
		Only property child nodes should be loaded.
	*/
	PROPERTY(false, false, true),
	
	/**
		Child nodes should be loaded in the forward-loading direction.
		Property nodes should also be loaded.
	*/
	FORWARD(true, false, true), 
	
	/**
		Child nodes should be loaded in the reverse-loading direction.
		Property nodes should also be loaded.
	*/
	REVERSE(false, true, true), 
	
	/*
		Child nodes should be loaded in the all directions. This 
		is usually only used on the root node.
	*/
	ALL(true, true, true);
	
	private final boolean loadForward;
	private final boolean loadReverse;
	private final boolean loadProperty;
	
	private ChildNodeDirection(boolean loadForward, boolean loadReverse, boolean loadProperty)
	{
		this.loadForward = loadForward;
		this.loadReverse = loadReverse;
		this.loadProperty = loadProperty;
	}
	
	/**
		@return TRUE if child nodes in the reverse-loading 
			direction should be loaded
	*/
	public boolean loadReverse()
	{
		return loadReverse;
	}

	/**
		@return TRUE if child nodes in the forward-loading 
			direction should be loaded
	*/
	public boolean loadForward()
	{
		return loadForward;
	}
	
	/**
		@return TRUE if property child nodes should be loaded
	*/
	public boolean loadProperty()
	{
		return loadProperty;
	}
	
}
