/**
	$Id: SoftwareValidator.java 5814 2011-10-18 11:46:59Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.overview.validator;

import net.sf.basedb.core.BasicItem;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.Item;
import net.sf.basedb.core.ItemSubtype;
import net.sf.basedb.core.PermissionDeniedException;
import net.sf.basedb.core.Project;
import net.sf.basedb.core.Software;
import net.sf.basedb.core.Subtypable;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.util.overview.Fix;
import net.sf.basedb.util.overview.OverviewContext;
import net.sf.basedb.util.overview.Validator;
import net.sf.basedb.util.overview.Node;

/**
	Validator implementation for software. Validation rules:
	<ul>
	<li>Missing item: {@link Validator#MISSING_SOFTWARE}
	<li>Access denied: {@link Validator#DENIED_SOFTWARE}
	<li>Incorrect software type: {@link Validator#INCORRECT_SOFTWARETYPE} 
	<li>Non-default software: {@link Validator#NONDEFAULT_SOFTWARE}
	</ul>

	@author Nicklas
	@version 2.10
	@base.modified $Date: 2011-10-18 13:46:59 +0200 (ti, 18 okt 2011) $
*/
public class SoftwareValidator
	extends NameableNodeValidator<Software>
{
	
	public SoftwareValidator()
	{
		super(Validator.MISSING_SOFTWARE, Validator.DENIED_SOFTWARE);
	}

	
	/* 
		From BasicValidator class
		-------------------------
	*/
	/**
		If the parent item has a subtype that is related to a SOFTWARE subtype, report
		the missing item as a {@link Validator#MISSING_SOFTWARE} failure. Otherwise,
		ignore the missing software.
		@return Always false
		@since 3.0
	*/
	@Override
	public boolean preMissingItem(DbControl dc, OverviewContext context, Node parentNode)
	{
		BasicItem parentItem = parentNode.getItem(dc);
		ItemSubtype expectedSubtype = getExpectedSoftwareSubtype(dc, parentItem);
		if (expectedSubtype != null)
		{
			context.createFailure(Validator.MISSING_SOFTWARE, parentNode, 
					"Missing software: " + expectedSubtype.getName(), getMissingItemFix(dc, parentNode));
		}
		return false;
	}


	/**
		Checks if the software is of the correct software type
		and, if a project is active, the software is the same as
		the default for the project.
	*/
	@Override
	public void postValidate(DbControl dc, OverviewContext context, Node node, Node parentNode)
	{
		super.postValidate(dc, context, node, parentNode);
		Software software = (Software)node.getItem(dc);
		
		// Validate the subtype of the software
		BasicItem parentItem = parentNode.getItem(dc);
		ItemSubtype expectedSubtype = getExpectedSoftwareSubtype(dc, parentItem);
		try
		{
			ItemSubtype subtype = software.getItemSubtype();
			if (expectedSubtype != null && !expectedSubtype.equals(subtype))
			{
				ItemSubtype mySubtype = ((Subtypable)parentItem).getItemSubtype();
				Fix subtypeFix = subtype == null ? null : 
					new Fix("Change related software subtype of '" + mySubtype.getName() + 
						"' to '" + subtype.getName() + "'", mySubtype);

				context.createFailure(Validator.INCORRECT_SOFTWARETYPE, parentNode, 
					"Expected a software with subtype: " + expectedSubtype.getName(),
					new Fix("Change software of '" + parentNode.getTitle() + "'", parentItem),
					new Fix("Change software type of '" + software.getName() + "' to '" + expectedSubtype.getName() + "'", software),
					subtypeFix
				);
			}
		}
		catch (PermissionDeniedException ex)
		{}
		
		// Check if using project default software
		Project project = context.getProject();
		if (project != null)
		{
			if (!project.isDefaultItem(software) && project.hasDefaultForItem(software))
			{
				context.createFailure(Validator.NONDEFAULT_SOFTWARE, parentNode, null,
					new Fix("Change software of '" + parentNode.getTitle() + "'", parentItem),
					new Fix("Add '" + software.getName() + "' as default software to project", project)
				);
			}
		}
	}

	/**
		@return Suggested fix is to add a software to the parent item
	*/
	@Override
	protected Fix getMissingItemFix(DbControl dc, Node parentNode)
	{
		BasicItem parentItem = parentNode.getItem();		
		return new Fix("Add software to '" + parentNode.getTitle() + "'", parentItem);
	}
	// --------------------------------
	
	/**
		Get the subtype of the associated software that we expect for the given
		parent item.
	*/
	protected ItemSubtype getExpectedSoftwareSubtype(DbControl dc, BasicItem parentItem)
	{
		ItemSubtype expectedSubtype = null;
		int defaultSubtypeForItemType = SystemItems.getId(Software.getDefaultSystemId(parentItem.getType()));
		try
		{
			if (parentItem instanceof Subtypable)
			{
				expectedSubtype = ItemSubtype.getRelatedSubtype(dc, 
						(Subtypable)parentItem, Item.SOFTWARE, defaultSubtypeForItemType);
			}
			else if (defaultSubtypeForItemType != 0)
			{
				expectedSubtype = ItemSubtype.getById(dc, defaultSubtypeForItemType);
			}
		}
		catch (PermissionDeniedException ex)
		{}
		return expectedSubtype;
	}
	
}
