/**
	$Id: UnitCache.java 7381 2017-05-22 12:35:27Z nicklas $

	Copyright (C) 2008 Nicklas Nordborg

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.util.units;

import java.util.HashMap;
import java.util.List;
import java.util.Map;

import net.sf.basedb.core.ConnectionClosedException;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemQuery;
import net.sf.basedb.core.Quantity;
import net.sf.basedb.core.SystemItems;
import net.sf.basedb.core.Type;
import net.sf.basedb.core.Unit;
import net.sf.basedb.core.query.Expressions;
import net.sf.basedb.core.query.Hql;
import net.sf.basedb.core.query.Restrictions;

/**
	Helper class for optimized lookup of units by symbol. 
	This class will cache units by their symbol and quantity,
	making repeated lookup for the same unit only hit the database 
	the first time. 
	<p>
	Note! The implementation of the cache is not thread-safe. If
	multiple threads are going to use the cache at the same time
	they must synchronize externally.
	
	@author Nicklas
	@version 2.9
	@base.modified $Date: 2017-05-22 14:35:27 +0200 (må, 22 maj 2017) $
*/
public class UnitCache 
{
	private DbControl dc;
	private ItemQuery<Unit> query;
	private Map<CacheKey, Unit> units;
	private boolean autoReload;
	
	/**
		Create a new unit cache. Before it can be used a 
		{@link DbControl} must be set by {@link #setDbControl(DbControl)}.		
	*/
	public UnitCache()
	{
		this(null);
	}
	
	/**
		Create a new unit cache which is ready to be used immediately.
		@param dc The DbControl to use for database access
	*/
	public UnitCache(DbControl dc)
	{
		this.dc = dc;
		this.units = new HashMap<CacheKey, Unit>();
	}
	
	/**
		Set the DbControl to use for database access
	*/
	public void setDbControl(DbControl dc)
	{
		this.dc = dc;
		this.query = null;
	}
	
	/**
		Get the DbControl that is currently used for database access.
	*/
	public DbControl getDbControl()
	{
		return dc;
	}

	/**
		Clear the cache of all units.
	*/
	public void resetCache()
	{
		units.clear();
	}

	
	/**
		Specifies if a cached unit that may have been loaded
		by a closed DbControl should be reloaded with the
		current DbControl or not. This assumes that the cache is
		currently using a connected DbControl. If not, a
		{@link ConnectionClosedException} is thrown when trying to
		find a unit. The default value for this setting is FALSE.
		@param autoReload TRUE to automatically reload the unit with
			the current DbControl
	*/
	public void setAutoReload(boolean autoReload)
	{
		this.autoReload = autoReload;
	}
	
	/**
		If units that are no longer connected to a DbControl should
		automatcally by reloaded with the current DbControl or not.
	*/
	public boolean isAutoReload()
	{
		return autoReload;
	}
	
	/**
		Finds a unit with a given symbol for a given quantity. 
		
		@param quantityId The ID of the quantity
		@param symbol The symbol of the unit
		@return A unit object, or null if no unit is found
	*/
	public Unit findUnit(int quantityId, String symbol)
	{
		CacheKey key = new CacheKey(quantityId, symbol);
		Unit unit = null;
		if (units.containsKey(key))
		{
			unit = units.get(key);
			if (autoReload && unit.isDetached())
			{
				if (dc == null || dc.isClosed())
				{
					throw new ConnectionClosedException();
				}
				unit = Unit.getById(dc, unit.getId());
				units.put(key, unit);
			}
		}
		else
		{
			unit = lookupUnitInDb(key, true);
		}
		return unit;
		
	}

	/**
		Finds a unit with a given symbol for a given quantity. 
		
		@param quantityId The system ID of the quantity
		@param symbol The symbol of the unit
		@return A unit object, or null if no unit is found
	*/
	public Unit findUnit(String quantityId, String symbol)
	{
		return findUnit(SystemItems.getId(quantityId), symbol);
	}
	/**
		Finds a unit with a given symbol for a given quantity. 
		
		@param quantity The quantity
		@param symbol The symbol of the unit
		@return A unit object, or null if no unit is found
	*/
	public Unit findUnit(Quantity quantity, String symbol)
	{
		return findUnit(quantity.getId(), symbol);
	}
	
	/**
		Lookup a unit in the database and optionally add it
		to the cache.
	*/
	private Unit lookupUnitInDb(CacheKey key, boolean addToCache)
	{
		if (dc == null || dc.isClosed()) throw new ConnectionClosedException();
		if (query == null)
		{
			query = Unit.getQuery();
			query.join(Hql.innerJoin("symbols", "smb"));
			query.restrict(
				Restrictions.eq(
					Hql.property("quantity"), Expressions.parameter("quantity")
				)
			);
			query.restrict(
				Restrictions.eq(
					Hql.property("smb", "symbol"), Expressions.parameter("symbol")
				)
			);
		}
		query.setParameter("quantity", key.quantityId, Type.INT);
		query.setParameter("symbol", key.symbol, Type.STRING);
		query.setMaxResults(1);
		List<Unit> list = query.list(dc);
		Unit unit = list.isEmpty() ? null : list.get(0);
		if (addToCache) units.put(key, unit);
		return unit;
	}
	
	/**
		Units are cached by (quantity,symbol) combination. This
		class is used as keys to the cache and overrides the 
		equals and hashcode methods to makes sure that 
		two keys with the same quantity and symbol are considered 
		equal.
		<p>
		This class also implements the comparable interface so that 
		it is possible to sort units in the order of their symbols.
		If two symbols are equal the one with the lowes quantity id
		is sorted first.
	*/
	static class CacheKey
		implements Comparable<CacheKey>
	{
		private final int quantityId;
		private final String symbol;

		CacheKey(int quantityId, String symbol)
		{
			this.quantityId = quantityId;
			this.symbol = symbol;
		}
		
		/**
			From the Object class
			---------------------
		*/
		@Override
		public int hashCode()
		{
			return symbol.hashCode();
		}
		
		@Override
		public boolean equals(Object other)
		{
			if (this == other) return true;
			if (other == null || other.getClass() != this.getClass()) return false;
			CacheKey ck = (CacheKey)other;
			return ck.quantityId == this.quantityId && ck.symbol.equals(this.symbol);
		}
		// -----------------------------
		
		/*
			From the Comparable interface
			-----------------------------
		*/
		@Override
		public int compareTo(CacheKey o) 
		{
			int result = this.symbol.compareTo(o.symbol);
			if (result == 0)
			{
				result = o.quantityId - this.quantityId;
			}
			return result;
		}
		// ------------------------------

	}
	
}
