/*
	$Id: Webclient.java 7403 2017-10-04 07:38:59Z nicklas $

	Copyright (C) 2005 Nicklas Nordborg
	Copyright (C) 2006 Jari Häkkinen, Nicklas Nordborg, Martin Svensson
	Copyright (C) 2007 Johan Enell, Nicklas Nordborg, Martin Svensson

	This file is part of BASE - BioArray Software Environment.
	Available at http://base.thep.lu.se/

	BASE is free software; you can redistribute it and/or
	modify it under the terms of the GNU General Public License
	as published by the Free Software Foundation; either version 3
	of the License, or (at your option) any later version.

	BASE is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with BASE. If not, see <http://www.gnu.org/licenses/>.
*/
package net.sf.basedb.install;

import net.sf.basedb.core.Config;
import net.sf.basedb.core.SessionControl;
import net.sf.basedb.clients.web.Base;
import net.sf.basedb.core.Application;
import net.sf.basedb.core.DbControl;
import net.sf.basedb.core.ItemKey;
import net.sf.basedb.core.Permission;
import net.sf.basedb.core.Client;
import net.sf.basedb.core.ClientDefaultSetting;
import net.sf.basedb.core.PluginDefinition;
import net.sf.basedb.core.BaseException;
import net.sf.basedb.core.ItemNotFoundException;
import net.sf.basedb.core.authentication.LoginRequest;
import net.sf.basedb.plugins.HelpImporter;
import net.sf.basedb.util.extensions.xml.PluginInfo;
import net.sf.basedb.util.extensions.xml.XmlLoader;

import java.io.InputStream;
import java.net.URL;

/**
	This class installs the web client application. This class
	can be run as a standalone application, but it is recommended
	that the {@link InitDB} class is used, since it will also
	install this class.

	@author Nicklas
	@version 2.0
	@see InitDB
*/
public final class Webclient
{

	/**
		This program requires two parameters: a user account and a password
		The user must have enough permissions to create a new client
		application and share it to everyone.
	 	@param args a String array with user account and password. 
	*/
	public static void main(String[] args)
	{
		if (args.length != 2)
		{
			showUsage();
			return;
		}
		String login = args[0];
		String password = args[1];
		try
		{
			install(login, password, null, null);
		}
		catch (Exception ex)
		{
			ex.printStackTrace();
		}
	}

	private static void showUsage()
	{
		System.out.println("Usage: installweb <login> <password>");
		System.out.println("       login: The login");
		System.out.println("       password: The password");
	}

	static void install(String login, String password, String adminName, String adminEmail)
		throws BaseException
	{
		Config.setProperty("extensions.disabled", "true");
		Application.start(false);
		SessionControl sc = Application.newSessionControl(null, null, null);
		LoginRequest loginRequest = new LoginRequest(login, password);
		loginRequest.setComment("Installing web client");
		sc.login(loginRequest);
		
		DbControl dc = sc.newDbControl();
		try
		{
			ItemKey shareToEveryone = ItemKey.getShareToEveryOneKey(dc, Permission.USE);
			
			Client client = null;
			try
			{
				client = Client.getByExternalId(dc, Base.WEBCLIENT_ID);
			}
			catch (ItemNotFoundException ex)
			{
				client = Client.getNew(dc, Base.WEBCLIENT_ID);
				client.setName("Web client");
				client.setDescription("The web interface to BASE");
				client.setSupportsDeviceVerification(true);
				client.setItemKey(shareToEveryone);
				dc.saveItem(client);
			}
	
			String latinText = "Fusce consectetuer. Suspendisse consectetuer wisi vitae eros. Quisque sapien sapien, "+
				"lobortis a, congue sed, tempus vitae, sapien. Aliquam rhoncus ultricies pede. Fusce "+
				"gravida pharetra tellus. Nulla tortor. Donec eu enim. Vivamus blandit. Pellentesque "+
				"habitant morbi tristique senectus et netus et malesuada fames ac turpis egestas. Mauris "+
				"pulvinar arcu id libero. Vestibulum pharetra. Proin id felis pharetra mauris feugiat adipiscing. "+
				"Cras auctor. Nullam turpis. Cras dignissim venenatis erat.<br>\n"+
				"[<b>The server administrator should replace this text, or remove it to disable this feature</b>]";
	
			createSetting(client, "server.about", latinText);
			createSetting(client, "server.admin.name", adminName);
			createSetting(client, "server.admin.email", adminEmail);
			createSetting(client, "server.forgotten.password", latinText);
			createSetting(client, "server.get.account", latinText);
			createSetting(client, "server.transferRate", Integer.toString(100 * 1024 * 1024)); // 100MB/s
			createSetting(client, "server.transferRate.download", Integer.toString(0)); // No limit
			createSetting(client, "server.links.reportbug", "http://base.thep.lu.se/#Feedback");
			createSetting(client, "server.links.help", "http://base.thep.lu.se/chrome/site/latest/html/index.html");
			createSetting(client, "server.links.faq", "");
			
			URL fileURL = Webclient.class.getResource("/web-extensions.xml");
			
			XmlLoader loader = new XmlLoader();
			try
			{
				InputStream in = fileURL.openStream();
				loader.loadXmlFile(in, "/web-extensions.xml", null, false);
			}
			catch (Exception ex)
			{
				throw new BaseException(ex);
			}
			
			for (PluginInfo info : loader.getPluginDefinitions())
			{
				PluginDefinition.installOrUpdate(dc, info, null, shareToEveryone);
			}
			dc.commit();
			installHelpTexts(sc, client);
		}
		finally
		{
			if (dc != null) dc.close();
		}
		
	}
	
	
	private static ClientDefaultSetting createSetting(Client client, String name, String value)
	{
		ClientDefaultSetting setting = null;
		if (client.isInDatabase()) 
		{
			try
			{
				setting = client.getDefaultSetting(name);
			}
			catch (ItemNotFoundException ex)
			{}
		}
		if (setting == null)
		{
			setting = client.newDefaultSetting(name, value);
			client.getDbControl().saveItem(setting);
		}
		else
		{
			String oldValue = setting.getValue();
			if (oldValue == null || "".equals(oldValue))
			{
				setting.setValue(value);
			}
		}
		return setting;
	}
	
	private static void installHelpTexts(SessionControl sc, Client client)
	{		
		String filePath = "/helptexts.xml";
		URL file = Webclient.class.getResource(filePath);
		InputStream is = Webclient.class.getResourceAsStream(filePath);
		if(is != null)
		{
			HelpImporter helpImp = new HelpImporter();
			helpImp.init(sc, null, null);
			helpImp.installHelpTexts(client, Config.getBoolean("helptext.update"), is, file.toString());
		}		
	}

}

